#
# TableView.py -- Table viewer for Ginga
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from collections import OrderedDict
import os.path
import numbers

import numpy as np

from ginga.util.viewer import ViewerBase
from ginga.util.paths import icondir
from ginga.table import AstroTable
from ginga.util.io import io_rgb


class TableViewBase(ViewerBase):
    """An abstract base class for displaying tables represented by
    astropy table objects.

    Parameters
    ----------
    logger : :py:class:`~logging.Logger` or `None`
        Logger for tracing and debugging. If not given, one will be created.

    settings : `~ginga.misc.Settings.SettingGroup` or `None`
        Viewer preferences. If not given, one will be created.

    """

    vname = 'Ginga Table'
    vtypes = [AstroTable.AstroTable]

    @classmethod
    def viewable(cls, dataobj):
        """Test whether `dataobj` is viewable by this viewer."""
        if not isinstance(dataobj, AstroTable.AstroTable):
            return False
        return True

    def __init__(self, logger=None, settings=None):
        ViewerBase.__init__(self, logger=logger, settings=settings)

        self._table = None

        self.settings.add_defaults(color_alternate_rows=True,
                                   max_rows_for_col_resize=5000)

        # no specific UI modes for this viewer
        self.set_allowed_modes([])

        # For callbacks
        for name in ('table-set', 'configure', ):
            self.enable_callback(name)

    def set_table(self, table):
        if not isinstance(table, AstroTable.AstroTable):
            raise ValueError("Wrong type of object to load: %s" % (
                str(type(table))))

        self._table = table

        self.make_callback('table-set', table)

    def get_table(self):
        return self._table

    # for compatibility with other Ginga viewers
    get_dataobj = get_table
    set_dataobj = set_table


class TableViewGw(TableViewBase):
    """A Ginga viewer for displaying tables of FITS data.
    """

    def __init__(self, logger=None, settings=None):
        super(TableViewGw, self).__init__(logger=logger, settings=settings)

        self.add_callback('table-set', self.set_table_cb)

        from ginga.gw import Widgets

        # Create the viewer as a Treeview widget
        color_alternate = self.settings.get('color_alternate_rows', True)
        self.widget = Widgets.TreeView(auto_expand=True,
                                       sortable=True,
                                       use_alt_row_color=color_alternate)

        rgb_opener = io_rgb.RGBFileHandler(self.logger)
        tmp_path = os.path.join(icondir, 'fits.png')
        placeholder_image = rgb_opener.load_file(tmp_path)
        self._rgb_array_placeholder = placeholder_image.get_data()

    def get_widget(self):
        return self.widget

    def _get_table_type(self, val):
        if isinstance(val, numbers.Number):
            if isinstance(val, int):
                return 'int'
            else:
                return 'float'
        elif isinstance(val, bool) or val in [np.True_, np.False_]:
            return 'bool'
        else:
            return 'str'

    def set_table_cb(self, viewer, table):
        """Display the given table object."""
        self.clear()
        tree_dict = OrderedDict()

        # Extract data as astropy table
        a_tab = table.get_data()

        columns = [('Row', '_DISPLAY_ROW', 'str')]

        # This is to get around some table widget implementations
        # not sorting numbers properly
        i_fmt = '{{0:0{0}d}}'.format(len(str(table.rows)))

        if table.kind == 'table-astropy':
            # Fill masked values, if applicable
            try:
                a_tab = a_tab.filled()
            except Exception:  # Just use original table
                pass

            # Table header with units
            row = a_tab[0]
            for i, c in enumerate(a_tab.columns.values()):
                col_str = '{0:^s}\n{1:^s}'.format(c.name, str(c.unit)) \
                    if c.unit is not None else '{0:^s}'.format(c.name)
                _data_type = self._get_table_type(row[i])
                columns.append((col_str, c.name, _data_type))

            # Table contents
            for i, row in enumerate(a_tab, 1):
                row_dct = dict(zip(row.colnames, row.as_void()))
                i_str = i_fmt.format(i)
                row_dct['_DISPLAY_ROW'] = i_str
                tree_dict[i_str] = row_dct

        elif table.kind == 'table-fitsio':
            colnames = table.colnames

            # Table header
            row = a_tab[0]
            for c_name in colnames:
                col_str = '{0:^s}'.format(c_name)
                _data_type = self._get_table_type(row[i])
                columns.append((col_str, c_name, _data_type))

            # Table contents
            for i, row in enumerate(a_tab, 1):
                row_dct = dict(zip(colnames, row))
                i_str = i_fmt.format(i)
                row_dct['_DISPLAY_ROW'] = i_str
                tree_dict[i_str] = row_dct

        else:
            raise ValueError(f"I don't know how to display tables of type '{table.kind}'")

        self.widget.setup_table(columns, 1, '_DISPLAY_ROW')
        self.widget.set_tree(tree_dict)

        # Resize column widths
        n_rows = len(tree_dict)
        if n_rows < self.settings.get('max_rows_for_col_resize', 5000):
            self.widget.set_optimal_column_widths()
            self.logger.debug('Resized columns for {0} row(s)'.format(n_rows))

        tablename = table.get('name', 'NoName')
        self.logger.debug('Displayed {0}'.format(tablename))

    def clear(self):
        self.widget.clear()

    def get_data_size(self):
        return (1, 1)

    def get_last_data_xy(self):
        return (0, 0)

    def get_window_size(self):
        return self.widget.get_size()

    def get_rgb_array(self):
        if not hasattr(self.widget, 'get_rgb_array'):
            return self._rgb_array_placeholder

        return self.widget.get_rgb_array()



#
# GwMain.py -- application threading help routines.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
import traceback
import threading
import time
import asyncio

from ginga.misc import Task, Future, Callback
from ginga.misc.aio_compat import get_event_loop as aio_get_event_loop
from ginga.toolkit import toolkit
from ginga.misc import log
from collections import deque

import queue as Queue


class GwMain(Callback.Callbacks):

    def __init__(self, queue=None, logger=None, ev_quit=None, app=None,
                 thread_pool=None, task_pool=None, async_mode=None):
        Callback.Callbacks.__init__(self)

        self.enable_callback('shutdown')

        # You can pass in a queue if you prefer to do so
        if not queue:
            queue = Queue.Queue()
        self.gui_queue = queue
        self.priority_gui_queue = Queue.PriorityQueue()

        # You can pass in a logger if you prefer to do so
        if logger is None:
            logger = log.NullLogger()
        self.logger = logger

        if not ev_quit:
            ev_quit = threading.Event()
        self.ev_quit = ev_quit

        self.app = app
        # Mark our thread id
        self.gui_thread_id = threading.get_ident()

        # ``task_pool`` is the preferred name; ``thread_pool`` is kept for
        # backward compatibility.  The pool need only implement the
        # addTask/startall/stopall interface (see ginga.misc.Task).
        if task_pool is None:
            task_pool = thread_pool
        self.threadPool = task_pool

        # In "async mode" the whole application runs on a single event
        # loop (e.g. Pyodide/the browser): there is no separate GUI thread,
        # so cross-thread marshaling and blocking waits must be avoided.
        # If not specified, infer it from the pool type.
        if async_mode is None:
            async_mode = isinstance(task_pool, Task.AsyncTaskPool)
        self.async_mode = async_mode

        # handle for the periodic event pump used in async mode
        self._async_pump_interval = 0.05

        # For asynchronous tasks on the thread pool
        self.tag = 'master'
        self.shares = ['threadPool', 'logger']

        self.oneshots = {}

    def get_widget(self):
        return self.app

    def get_threadPool(self):
        return self.threadPool

    def get_taskpool_type(self):
        """Return the kind of task pool in use.

        Returns
        -------
        kind : str
            ``'async'`` if tasks run on a single event loop (e.g. when
            running in-situ under Pyodide), or ``'thread'`` if they run
            on a pool of worker threads.

        This lets callers (e.g. plugins doing network I/O) choose an
        appropriate strategy -- a synchronous ``requests`` download on a
        worker thread vs. an awaited browser fetch -- without inspecting
        internals.
        """
        return 'async' if self.async_mode else 'thread'

    def set_gui_thread(self):
        self.gui_thread_id = threading.get_ident()

    def _execute_future(self, future):
        # Execute the GUI method
        try:
            try:
                future.thaw(suppress_exception=False)

            except Exception as e:
                self.logger.error("gui event loop error: %s" % str(e))
                try:
                    (type, value, tb) = sys.exc_info()
                    tb_str = "".join(traceback.format_tb(tb))
                    self.logger.error("Traceback:\n%s" % (tb_str))

                except Exception:
                    self.logger.error("Traceback information unavailable.")

                future.resolve(e)

        except Exception as e2:
            self.logger.error("Exception resolving future: %s" % str(e2))

    def update_pending(self, timeout=0.0, elapsed_max=0.02):

        self.assert_gui_thread()

        # Process "out-of-band" events
        # self.logger.debug("1. processing out-of-band GUI events")
        try:
            self.app.process_events()

        except Exception as e:
            self.logger.error(str(e))

        # Process "in-band" GUI events
        # self.logger.debug("2. processing approx %d in-band GUI events" % (
        #                    self.gui_queue.qsize()))
        done = False
        time_start = time.time()

        # First process priority futures
        while not done:
            try:
                future = self.priority_gui_queue.get(block=False)
                self._execute_future(future)

            except Queue.Empty:
                break

            if time.time() - time_start > elapsed_max:
                done = True

        # Next process non-priority futures
        while not done:
            try:
                future = self.gui_queue.get(block=True,
                                            timeout=timeout)
                self._execute_future(future)

            except Queue.Empty:
                done = True

            if time.time() - time_start > elapsed_max:
                done = True

        # Execute all the one-shots
        # self.logger.debug("3. processing one-shot GUI events")
        deqs = list(filter(lambda deq: len(deq) > 0, self.oneshots.values()))
        for deq in deqs:
            try:
                future = deq.pop()

                self._execute_future(future)

            except IndexError:
                continue

        # Process "out-of-band" events, again
        # self.logger.debug("4. processing out-of-band GUI events")
        try:
            self.app.process_events()

        except Exception as e:
            self.logger.error(str(e))

        # self.logger.debug("5. done")

    def gui_do_priority(self, priority, method, *args, **kwdargs):
        """General method for asynchronously calling into the GUI.
        It makes a future to call the given (method) with the given (args)
        and (kwdargs) inside the gui thread.  If the calling thread is a
        non-gui thread the future is returned.
        """
        future = Future.Future(priority=priority)
        future.freeze(method, *args, **kwdargs)
        self.priority_gui_queue.put(future)

        my_id = threading.get_ident()
        if my_id != self.gui_thread_id:
            return future

    def gui_do(self, method, *args, **kwdargs):
        future = Future.Future(priority=0)
        future.freeze(method, *args, **kwdargs)
        self.gui_queue.put(future)

        my_id = threading.get_ident()
        if my_id != self.gui_thread_id:
            return future

    def gui_call(self, method, *args, **kwdargs):
        """General method for synchronously calling into the GUI.
        This waits until the method has completed before returning.
        """
        my_id = threading.get_ident()
        if my_id == self.gui_thread_id:
            return method(*args, **kwdargs)
        else:
            future = self.gui_do(method, *args, **kwdargs)
            return future.wait()

    def gui_do_future(self, future):
        self.gui_queue.put(future)
        return future

    def gui_do_oneshot(self, catname, method, *args, **kwdargs):
        if catname not in self.oneshots:
            deq = self.oneshots.setdefault(catname, deque([], 1))
        else:
            deq = self.oneshots[catname]

        future = Future.Future()
        future.freeze(method, *args, **kwdargs)
        deq.append(future)

        my_id = threading.get_ident()
        if my_id != self.gui_thread_id:
            return future

    def make_async_gui_callback(self, name, *args, **kwargs):
        # NOTE: asynchronous!
        self.gui_do(self.make_callback, name, *args, **kwargs)

    def make_gui_callback(self, name, *args, **kwargs):
        if self.is_gui_thread():
            return self.make_callback(name, *args, **kwargs)
        else:
            # note: this cannot be "gui_call"--locks viewer.
            # so call becomes async when a non-gui thread invokes it
            self.gui_do(self.make_callback, name, *args, **kwargs)

    def _make_func_task(self, method, args, kwdargs):
        # In async mode use a cooperative task that awaits coroutine
        # results on the single event loop; otherwise a plain FuncTask
        # for the thread pool.
        if self.async_mode:
            return Task.AsyncFuncTask(method, args, kwdargs, logger=self.logger)
        return Task.FuncTask(method, args, kwdargs, logger=self.logger)

    def nongui_do(self, method, *args, **kwdargs):
        task = self._make_func_task(method, args, kwdargs)
        return self.nongui_do_task(task)

    def nongui_do_cb(self, tup, method, *args, **kwdargs):
        task = self._make_func_task(method, args, kwdargs)
        _args = [] if len(tup) == 1 else tup[1:]
        task.add_callback('resolved', tup[0], *_args)
        return self.nongui_do_task(task)

    def nongui_do_future(self, future):
        task = Task.FuncTask(future.thaw, (), {}, logger=self.logger)
        return self.nongui_do_task(task)

    def nongui_do_task(self, task):
        try:
            task.init_and_start(self)
            return task
        except Exception as e:
            self.logger.error("Error starting task: %s" % (str(e)))
            raise e

    def nongui_foreach(self, items, worker, on_each=None, on_done=None,
                       yield_interval=0.05, is_cancelled=None):
        """Run ``worker(item)`` for each item off the GUI while keeping the
        UI responsive.

        In async (single event loop) mode this runs as a coroutine that
        yields control back to the loop on a time budget (see
        ``yield_interval``), so UI events and repaints are serviced as the
        work proceeds; in threaded mode it runs as a single nongui task on a
        worker thread.  The caller-facing API is identical in both modes.

        This is the cooperative escape hatch for CPU-bound "nongui" work in
        the single-threaded (browser/Pyodide) backend, where a nongui task
        is just a coroutine on the one event loop: without periodic yields a
        long loop starves the UI exactly like a synchronous call would.

        Parameters
        ----------
        items : iterable
            The work items.
        worker : callable
            ``worker(item) -> result``.  Runs off the GUI; must not touch
            widgets directly.
        on_each : callable or None
            ``on_each(item, result)``, scheduled on the GUI (via gui_do)
            after each item completes -- use for incremental display /
            progress.
        on_done : callable or None
            ``on_done(results)``, scheduled on the GUI after all items
            complete (not called if cancelled).
        yield_interval : float
            Minimum wall-clock seconds between event-loop yields (async mode
            only).  The loop is given control only after this much time has
            elapsed, *not* after every item -- each yield is a full
            round-trip to the browser event loop costing real time, so
            yielding per item would make the total cost roughly constant
            regardless of how fast the work itself is (e.g. when results are
            cached).  Yielding on a time budget keeps the UI responsive while
            letting a fast (cached) run finish with few yields.
        is_cancelled : callable or None
            ``is_cancelled() -> bool``, checked before each item; abort if
            it returns True.

        Returns the task object.
        """
        items = list(items)

        if self.async_mode:
            async def _run_async():
                results = []
                t_last = time.monotonic()
                for item in items:
                    if is_cancelled is not None and is_cancelled():
                        return
                    results.append(worker(item))
                    if on_each is not None:
                        self.gui_do(on_each, item, results[-1])
                    if time.monotonic() - t_last >= yield_interval:
                        await asyncio.sleep(0)
                        t_last = time.monotonic()
                if on_done is not None:
                    self.gui_do(on_done, results)
            return self.nongui_do(_run_async)

        def _run_thread():
            results = []
            for item in items:
                if is_cancelled is not None and is_cancelled():
                    return
                results.append(worker(item))
                if on_each is not None:
                    self.gui_do(on_each, item, results[-1])
            if on_done is not None:
                self.gui_do(on_done, results)
        return self.nongui_do(_run_thread)

    def is_gui_thread(self):
        # In async mode there is only one thread; treat it as the GUI
        # thread so gui_call() runs inline and never blocks on a future.
        if self.async_mode:
            return True
        my_id = threading.get_ident()
        return my_id == self.gui_thread_id

    def assert_gui_thread(self):
        if self.async_mode:
            return
        my_id = threading.get_ident()
        assert my_id == self.gui_thread_id, \
            Exception("Non-GUI thread (%d) is executing GUI (%d) code!" % (
                my_id, self.gui_thread_id))

    def assert_nongui_thread(self):
        # Single-threaded: everything runs on the one (GUI) thread, so
        # this assertion is not meaningful in async mode.
        if self.async_mode:
            return
        my_id = threading.get_ident()
        assert my_id != self.gui_thread_id, \
            Exception("GUI thread (%d) is executing non-GUI code!" % (
                my_id))

    def mainloop(self, timeout=0.001):
        # Mark our thread id
        self.gui_thread_id = threading.get_ident()

        if self.async_mode:
            # Single event loop.  Two sub-cases:
            #
            #  * A loop is already running (e.g. Pyodide/the browser): we
            #    cannot block, so install a cooperative pump and return --
            #    the host event loop keeps everything alive.
            #
            #  * No loop is running (e.g. a native CLI invocation with
            #    --task-pool=async): we own the loop and must run it here
            #    until quit, otherwise mainloop() would return immediately
            #    and the application would exit early.
            try:
                asyncio.get_running_loop()
                hosted = True
            except RuntimeError:
                hosted = False

            if hosted:
                self.start_async_pump()
                return

            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            self.start_async_pump()
            try:
                loop.run_until_complete(self._async_wait_quit())
            finally:
                loop.close()
            return

        if toolkit == 'gtk4':
            self.app.add_periodic_callback(0.1,
                                           lambda: self.update_pending(timeout=timeout))
            self.app._mainloop()
            return

        while not self.ev_quit.is_set():
            self.update_pending(timeout=timeout)

    def start_async_pump(self, interval=None):
        """Drive ``update_pending`` from a periodic callback on the running
        event loop, for single-threaded (async) operation.  Returns
        immediately; the host event loop continues running.
        """
        if interval is not None:
            self._async_pump_interval = interval
        loop = aio_get_event_loop()

        def _pump():
            if self.ev_quit.is_set():
                return
            try:
                # non-blocking: drain whatever is queued and return
                self.update_pending(timeout=0.0)
            except Exception as e:
                self.logger.error("error in event pump: %s" % (str(e),))
            loop.call_later(self._async_pump_interval, _pump)

        loop.call_soon(_pump)

    async def _async_wait_quit(self):
        """Keep the (owned) event loop alive until ``ev_quit`` is set.

        Used when running async-mode on a native platform, where we run
        the loop ourselves rather than relying on a host (browser) loop.
        """
        while not self.ev_quit.is_set():
            await asyncio.sleep(self._async_pump_interval)

    def gui_quit(self):
        """Call this to cause the GUI thread to quit the mainloop."""
        self.ev_quit.set()

        self.make_callback('shutdown')

        self.app.process_end()

    def _quit(self):
        self.gui_quit()

# END



#
# PgHelp.py -- web application threading help routines.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os
import binascii
from io import BytesIO

from PIL import Image

from ginga.fonts import font_asst
from ginga.util import icon_helper


def get_image_src_from_buffer(img_buf, imgtype='png'):
    if not isinstance(img_buf, bytes):
        img_buf = img_buf.encode('latin1')
    img_string = binascii.b2a_base64(img_buf)
    if isinstance(img_string, bytes):
        img_string = img_string.decode("utf-8")
    return f'data:image/{imgtype};base64,' + img_string


def get_icon(iconpath, size=None, format='svg'):

    if size is not None:
        wd, ht = size
    else:
        wd, ht = 24, 24

    if iconpath.lower().endswith('.svg'):
        if format.startswith('svg'):
            # TEMP?
            # Scalable Vector Graphics should be scalable on the browser side
            with open(iconpath, 'rb') as icon_f:
                svg_buf = icon_f.read()
                icon_uri = get_image_src_from_buffer(svg_buf, imgtype='svg+xml')
                return icon_uri

        if format == 'png':
            img_buf = icon_helper.load_svg_to_pngbuf(iconpath,
                                                     wd_px=wd, ht_px=ht)
            icon_uri = get_image_src_from_buffer(img_buf.getvalue(),
                                                 imgtype=format)
            return icon_uri

    # other types (raster images: png, gif, jpg, ...).
    if size is None:
        # no resize requested -- pass the original bytes straight through as
        # a data URI (already a browser-renderable format).
        ext = os.path.splitext(iconpath)[1].lower().lstrip('.') or 'png'
        imgtype = 'jpeg' if ext in ('jpg', 'jpeg') else ext
        with open(iconpath, 'rb') as icon_f:
            buf = icon_f.read()
        return get_image_src_from_buffer(buf, imgtype=imgtype)

    # resize to the requested (wd, ht).  Convert to RGBA first: resizing a
    # palette ('P') image and re-saving as 'P' can drop the transparent
    # color (rendering the icon blank), so work in RGBA to keep transparency.
    image = Image.open(iconpath).convert('RGBA')
    image = image.resize((wd, ht))
    img_buf = BytesIO()
    image.save(img_buf, format='PNG')
    return get_image_src_from_buffer(img_buf.getvalue(), imgtype='png')


def get_image(imgpath, size=None, format='png'):
    if size is not None:
        wd, ht = size
    else:
        wd, ht = 24, 24
    if imgpath.endswith('.svg'):
        # Scalable Vector Graphics
        img_buf = icon_helper.load_svg_to_pngbuf(imgpath, wd_px=wd, ht_px=ht)
        img = get_image_src_from_buffer(img_buf.getvalue(), imgtype=format)
        return img

    else:
        return get_icon(imgpath, size=size, format=format)


def get_font(font_spec, font_size):
    """Function to obtain a font for the Pg backend.

    Parameters
    ----------
    font_spec : str or `~ginga.fonts.font_asst.Font`
        The desired font

    font_size : int
        The point size requested for the given font

    Returns
    -------
    font : dict
        The desired font information in native backend form
    """
    key = ('pg', font_spec, font_size)
    try:
        return font_asst.get_cache(key)

    except KeyError:
        pass

    if isinstance(font_spec, str):
        font_tup = font_asst.parse_font(font_spec)
    elif isinstance(font_spec, font_asst.Font):
        font_tup = font_spec
    else:
        raise ValueError("not a valid font spec: {}".format(str(font_spec)))

    font_dct = font_tup._asdict()
    # emit a CSS font-family fallback list (e.g. '"ubuntu mono",
    # monospace') rather than a single family, so the browser can fall
    # back gracefully when the preferred face isn't available
    font_dct['family'] = font_asst.get_css_family_list(font_tup.family)
    font_dct['size'] = font_size
    # cache this dict for faster lookups hence
    font_asst.add_cache(key, font_dct)
    if isinstance(font_spec, str):
        # also store the font under a secondary key
        key2 = ('pg', font_tup, font_size)
        font_asst.add_cache(key2, font_dct)

    return font_dct


# END



#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""Regenerate the gettext message template (ginga.pot) for the reference
viewer.

Ginga marks translatable strings two ways, and this tool extracts both:

1. Explicit ``_tr("...")`` / ``_("...")`` calls (standard Babel extraction).
2. ``build_info()`` caption titles.  Those are plain strings inside caption
   tuples (``('Label:', 'label', ...)``); ``build_info`` translates the
   display text at runtime while keeping the widget key in English (see
   ``ginga.locale.localize.translate_caption``), so Babel cannot see them as
   ordinary keyword calls.  This module AST-parses caption tuples and pulls
   the titles of the user-visible widget types (skipping ``^``-opted-out
   titles and non-display types).

Usage::

    python -m ginga.locale._extract [srcdir] [-o outfile]

Requires Babel (a build/development dependency, not a runtime one).
"""
import ast
import os
import sys

from ginga.locale.localize import (
    KNOWN_WTYPES, TRANSLATABLE_WTYPES, NO_TRANSLATE_PREFIX)


def iter_caption_messages(source, filename='<unknown>'):
    """Yield ``(lineno, text)`` for each translatable ``build_info()`` caption
    title found in `source` (a string of Python code).
    """
    try:
        tree = ast.parse(source, filename=filename)
    except SyntaxError:
        return
    for node in ast.walk(tree):
        # A captions argument is a non-empty tuple/list whose every element is
        # itself a tuple (a caption row).  Requiring the tuple-of-tuples shape
        # avoids matching flat wtype tuples like ('checkbox', 'checkbutton')
        # that appear in the widget library itself.
        if not isinstance(node, (ast.Tuple, ast.List)):
            continue
        rows = node.elts
        if len(rows) == 0 or not all(isinstance(r, ast.Tuple) for r in rows):
            continue
        for row in rows:
            elts = row.elts
            n = len(elts)
            if n < 2 or (n % 2) != 0:
                continue
            # a caption row has a known wtype string at every odd position
            wtype_nodes = elts[1::2]
            if not all(isinstance(t, ast.Constant) and
                       isinstance(t.value, str) and
                       t.value in KNOWN_WTYPES for t in wtype_nodes):
                continue
            for i in range(0, n, 2):
                name_node, wtype = elts[i], elts[i + 1].value
                if wtype not in TRANSLATABLE_WTYPES:
                    continue
                if not (isinstance(name_node, ast.Constant) and
                        isinstance(name_node.value, str)):
                    continue
                title = name_node.value
                if title.startswith(NO_TRANSLATE_PREFIX) or title.strip() == '':
                    continue
                yield (name_node.lineno, title)


def build_catalog(srcdir='ginga'):
    from babel.messages.catalog import Catalog
    from babel.messages.extract import extract_from_dir

    catalog = Catalog(project='ginga', charset='utf-8')

    # pass 1: explicit _tr()/_()/N_() calls
    for filename, lineno, message, comments, context in extract_from_dir(
            srcdir, method_map=[('**.py', 'python')],
            keywords={'_tr': None, '_': None, 'N_': None}):
        catalog.add(message, locations=[(os.path.join(srcdir, filename),
                                         lineno)],
                    auto_comments=comments)

    # pass 2: build_info() caption titles
    for root, dirs, files in os.walk(srcdir):
        for fn in sorted(files):
            if not fn.endswith('.py'):
                continue
            path = os.path.join(root, fn)
            with open(path, encoding='utf-8') as f:
                src = f.read()
            for lineno, text in iter_caption_messages(src, path):
                catalog.add(text, locations=[(path, lineno)])

    return catalog


def main(argv=None):
    from babel.messages.pofile import write_po

    argv = list(sys.argv[1:] if argv is None else argv)
    outfile = 'ginga/locale/ginga.pot'
    if '-o' in argv:
        i = argv.index('-o')
        outfile = argv[i + 1]
        del argv[i:i + 2]
    srcdir = argv[0] if argv else 'ginga'

    catalog = build_catalog(srcdir)
    with open(outfile, 'wb') as f:
        write_po(f, catalog, width=76, sort_output=True)
    print("wrote %d messages to %s" % (len(catalog), outfile))
    return 0


if __name__ == '__main__':
    sys.exit(main())



#
# tablehelper.py -- feed ginga TableView widgets from astropy Tables / numpy
#
# The TableView wrappers (qtw / gtk3w / gtk4w / web.pgw) all build on
# set_columns() + set_rows(), so these converters just produce the column
# descriptors and row lists those methods already accept.
#
# astropy is intentionally NOT imported here: a Table is handled duck-typed
# (needs only ``.colnames`` and ``table[name]`` column access), so importing
# this module never drags in astropy.
#
import numpy as np

# numpy dtype.kind -> TableView column type (drives display/sort).  Anything
# unrecognised (datetime, object, bytes, ...) falls back to 'string'.
_KIND_TO_TYPE = {'i': 'int', 'u': 'int', 'f': 'float', 'b': 'bool'}


def _col_type(col):
    dt = getattr(col, 'dtype', None)
    return _KIND_TO_TYPE.get(getattr(dt, 'kind', None), 'string')


def _col_values(col):
    # native python list; masked entries come back as None
    if hasattr(col, 'tolist'):
        return col.tolist()
    return list(col)


def _colnames(table):
    # astropy Table exposes .colnames; pandas DataFrame exposes .columns.
    # (astropy also has a .columns mapping, so check .colnames first.)  This
    # is all duck-typed -- neither astropy nor pandas is imported here.
    names = getattr(table, 'colnames', None)
    if names is not None:
        return list(names)
    return list(table.columns)


def columns_from_table(table):
    """Build TableView column descriptors from an astropy Table or a pandas
    DataFrame.

    The column name becomes both the key and the label; the type is inferred
    from the column's dtype.  Duck-typed: works with anything exposing
    ``.colnames``/``.columns`` and ``table[name]``.
    """
    return [dict(label=name, key=name, type=_col_type(table[name]))
            for name in _colnames(table)]


def rows_from_table(table):
    """Return an astropy Table's (or pandas DataFrame's) contents as a list of
    row dicts keyed by column name (values converted to native python types)."""
    names = _colnames(table)
    cols = [_col_values(table[name]) for name in names]
    return [dict(zip(names, vals)) for vals in zip(*cols)]


def rows_from_ndarray(data):
    """Convert a numpy array into TableView rows.

    * structured / record array -> list of dicts keyed by field name
    * 2-D array -> list of positional rows (mapped to columns in order)
    * 1-D array -> list of single-value rows

    Values are converted to native python types (via ``tolist``), so numpy
    scalars don't leak into the widget / JSON layers.
    """
    if data.dtype.names is not None:
        names = data.dtype.names
        return [dict(zip(names, rec)) for rec in data.tolist()]
    if data.ndim <= 1:
        return [[v] for v in data.tolist()]
    return data.tolist()


def is_ndarray(data):
    """True if ``data`` looks like a numpy array (duck-typed on dtype+ndim),
    so callers can decide to route it through :func:`rows_from_ndarray`
    without importing numpy themselves."""
    return isinstance(data, np.ndarray)

#END



#
# Widgets.py -- wrapped HTML widgets and convenience functions
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from functools import reduce
import os.path

in_situ_web = False
try:
    from pgwidgets_js.pyodide import Widgets as PGW
    from pgwidgets_js.pyodide.widget import Widget
    # <-- we are being imported from pyodide/pyscript
    in_situ_web = True
except ImportError:
    # <-- we are being imported from python
    from pgwidgets.sync import Widgets as PGW
    from pgwidgets.sync.widget import Widget
    from pgwidgets.sync.application import Application as PGW_Application
    from pgwidgets.extras.file_browser import FileBrowser

from ginga.misc.Callback import Callbacks
from ginga.misc import Bunch, Settings
from ginga.web.pgw import PgHelp
from ginga.util.paths import icondir, app_icon_path
from ginga.fonts import font_asst
from ginga.locale.localize import translate_caption, _tr

__all__ = ['WidgetError', 'Widget', 'WidgetBase', 'TextEntry', 'TextEntrySet',
           'TextArea', 'TextSource', 'Dial', 'Label', 'HSeparator',
           'VSeparator', 'Button', 'ComboBox',
           'SpinBox', 'Slider', 'ScrollBar', 'CheckBox', 'ToggleButton',
           'RadioButton', 'Image', 'Canvas', 'ProgressBar', 'StatusBar',
           'TreeView', 'TableView', 'Box', 'HBox', 'VBox', 'ButtonBox',
           'FixedLayout', 'Frame', 'Expander', 'TabWidget', 'StackWidget',
           'MDIWidget', 'ScrollArea', 'Splitter', 'GridBox',
           'ToolbarAction', 'Toolbar', 'MenuAction', 'Menu', 'Menubar',
           'Page', 'TopLevel', 'Dialog', 'FileDialog', 'BrowserFileDialog',
           'ColorDialog', 'MessageDialog', 'Timer', 'Application',
           'name_mangle', 'make_widget', 'hadjust', 'build_info', 'wrap']


class WidgetError(Exception):
    """For errors thrown in this module."""
    pass


# reference to the created application
_app = None
# reference to the created session
_session_id = 1
_session = None


# BASE
class CallbackMixin(Callbacks):

    def __init__(self):
        Callbacks.__init__(self)

    def _enable_callback(self, name):
        if name not in self.cb:
            self.clear_callback(name)

    def _make_callback(self, name, *args, **kwargs):
        return Callbacks.make_callback(self, name, *args, **kwargs)

    def enable_callback(self, name):
        self._enable_callback(name)

    def has_callback(self, name):
        if name in self.cb:
            return True
        if not hasattr(self, '_defn') or self._defn is None:
            return False
        return name in self._defn.get("callbacks", [])

    def add_callback(self, name, cb_fn, *args, **kwargs):
        if Callbacks.has_callback(self, name):
            Callbacks.add_callback(self, name, cb_fn, *args, **kwargs)
        else:
            # ``Widget`` is the pgwidgets base class imported at module
            # top (sync or pyodide variant).  ``PGW`` is the *namespace*
            # of concrete widget classes and has no ``Widget`` attribute
            # in the in-situ/pyodide build, so use the imported class.
            Widget.add_callback(self, name, cb_fn, *args, **kwargs)


class WidgetMixin(CallbackMixin):

    def __init__(self):
        CallbackMixin.__init__(self)

        # for storing extra data that the user wants to associate
        # with a widget
        self.extdata = Bunch.Bunch()

    def get_app(self):
        return _app

    def get_widget(self):
        return self

    def get_size(self):
        res = super().get_size()
        if res is None:
            # pgwidgets can return None for the size of a widget
            # if it hasn't been realized
            return 10, 10
        return res

    def focus(self):
        # focus() => set_focus() in pgwidgets
        self.set_focus()

    def delete(self):
        # delete() => destroy() in pgwidgets
        self.destroy()

    def set_font(self, font, size=10):
        # The pgwidgets-js ``set_font`` is positional —
        # ``set_font(family, size, weight, style)`` setting the
        # corresponding CSS properties on the element.  Unpack the
        # Ginga-style font here so the JS side gets a CSS
        # family string rather than a dict object that stringifies
        # to "[object Object]" (which the browser then falls back
        # to its default serif font for).
        # PgHelp.get_font accepts a str spec or a font_asst.Font and returns
        # a CSS-style dict (family/size/weight/style); a dict is used as-is.
        font_info = font
        if isinstance(font_info, (str, font_asst.Font)):
            font_info = PgHelp.get_font(font_info, size)
        family = font_info.get('family', 'sans-serif')
        pts = font_info.get('size', size)
        weight = font_info.get('weight', 'normal')
        style = font_info.get('style', 'normal')
        super().set_font(family, pts, weight, style)

    def set_margins(self, l, r, t, b):
        # used interchangably with the idea of border width in Ginga
        # widgets
        self.set_padding([l, r, t, b])

    def set_border_width(self, px):
        self.set_padding(px)

    def cfg_expand(self, horizontal='fixed', vertical='fixed'):
        # this is for compatibility with Qt widgets
        pass


class ContainerWidgetMixin(WidgetMixin):

    def __init__(self):
        WidgetMixin.__init__(self)

        self._enable_callback('widget-removed')
        # Python-side list of child *wrappers*.  The JS side only knows the
        # JS widgets, so its get_children() returns JsProxy objects with no
        # get_widget()/extdata; ginga (e.g. Desktop.get_configuration)
        # expects the ginga wrappers, like the qt/gtk backends.  Track them
        # here so get_children() can return them.
        self.children = []

    def add_widget(self, child, *args, **kwargs):
        self.add_ref(child)
        return super().add_widget(child, *args, **kwargs)

    def add_ref(self, ref):
        self.children.append(ref)

    def get_children(self):
        return self.children

    def num_children(self):
        return len(self.children)

    def remove(self, child, delete=False):
        super().remove(child, destroy=delete)
        if child in self.children:
            self.children.remove(child)
        self._make_callback('widget-removed', child)

    def remove_all(self, delete=False):
        super().remove_all(destroy=delete)
        self.children = []


# for compatibility
class WidgetBase(WidgetMixin):
    pass


# NOTE (parked): assigning ``WidgetBase = WidgetMixin`` makes Desktop's
# ``isinstance(widget, WidgetBase)`` checks match pg widgets, so layout
# width/height size directives get applied to the top-level content.
# However that currently leaves the TopLevel content not resizing with
# the window, so it is disabled for now.  Revisit together with the
# pgwidgets-js TopLevel / get_size sizing work.
# WidgetBase = WidgetMixin


class ApplicationBase(Callbacks):
    """Base application class for Ginga-wrapped PG Widgets

    This is subclassed depending on whether we are running completely
    within the browser (Pyodide/PyScript) or externally using the web
    socket interface (pgwidgets-python).
    """
    def __init__(self, logger=None, host='localhost', port=9909, ws_port=None,
                 settings=None):
        global _app
        Callbacks.__init__(self)
        self.logger = logger

        # module-level assignments!
        self._windows = []
        _app = self

        if settings is None:
            settings = Settings.SettingGroup(logger=self.logger)
        self.settings = settings

        for name in ['close', 'shutdown']:
            self.enable_callback(name)

    def get_screen_size(self):
        # subclass should override this
        return (1000, 1000)

    def get_url(self):
        # subclass should override this
        return None

    def is_web_backend(self):
        return True

    def is_in_situ(self):
        """Return True when running *in-situ* in the browser (Pyodide/
        PyScript) -- i.e. Python itself is executing in the browser --
        as opposed to driving a remote browser over a web socket.

        Combine with is_web_backend() to decide, e.g., whether network
        I/O must go through the browser (subject to CORS / needing a
        proxy) vs. a normal server-side request.
        """
        return in_situ_web

    def make_window(self, title=None):
        win = TopLevel(title=title, moveable=True, resizable=True)
        self.add_window(win)
        return win

    def add_window(self, win):
        if win not in self._windows:
            self._windows.append(win)

    def remove_window(self, win):
        if win in self._windows:
            self._windows.remove(win)

    def make_timer(self):
        return Timer()

    def process_events(self, timeout=0.0):
        # In the browser (in-situ) there is nothing to pump: the JS event
        # loop dispatches widget events directly.  Provided so the GUI
        # event pump (GwMain.update_pending) can call it unconditionally.
        # The websocket Application overrides this with a real impl.
        pass

    def close(self):
        """Called when someone is asking the application to close.
        Can register for this callback if you want an application-wide
        event to confirm closure.
        """
        self.make_callback('close')

    def quit(self):
        """Called when someone is forcibly quitting the application.
        Can register for this callback if you want an application-wide
        event to clean up before shutdown.
        """
        self.make_callback('shutdown')

        super().close()

    def mainloop(self, timeout=None):
        self.start()
        self.run()


if in_situ_web:
    # <-- Pyodide/Pyscript

    class Application(ApplicationBase):
        """Application class when we are running *in-situ* in the browser."""

        def __init__(self, logger=None, host='localhost', port=9909,
                     ws_port=None, ws_sock=None, settings=None, token=None):
            ApplicationBase.__init__(self, logger=logger, host=host, port=port,
                                     ws_port=ws_port, settings=settings)

        def open_url(self, url):
            """Open *url* in a new browser tab.  We are *in* the browser,
            so call window.open directly."""
            from js import window
            window.open(url, '_blank')

    def register_font(self, family, path, weight='normal', style='normal'):
        # no-op for now
        pass

    def set_default_font(self, family, size=None, weight=None, style=None):
        font_asst.add_alias('sans', family.lower())

    FileBrowser = PGW.FileDialog  # noqa

else:
    # <-- Running on a host and using the web socket interface

    class Application(PGW_Application, ApplicationBase):
        """Application class when only the GUI is in the browser."""

        def __init__(self, logger=None, host='localhost', port=9909,
                     http_server=None, ws_port=None, ws_sock=None,
                     max_sessions=1, token=None, settings=None):
            global _session
            ApplicationBase.__init__(self, logger=logger, host=host, port=port,
                                     ws_port=ws_port, settings=settings)
            if ws_port is None:
                ws_port = port + 1
            if http_server is None:
                http_server = self.settings.get('http_server', False)
            # NOTE: set unconditionally so get_url()/mainloop() work even
            # when ``http_server`` is passed explicitly (not via settings)
            self.use_http_server = http_server
            if token is None:
                token = self.settings.get('token', None)

            PGW_Application.__init__(self, ws_port=ws_port, ws_sock=ws_sock,
                                     http_port=port, host=host,
                                     http_server=http_server,
                                     max_sessions=max_sessions,
                                     concurrency_handling='serialized')

            # we create a default session that is used by the widgets
            self.default_session = self.create_session(session_id=_session_id,
                                                       token=token)
            _session = self.default_session

            if self.settings.get('load_bundled_fonts', False):
                # add bundled fonts
                for key in font_asst.get_loadable_fonts():
                    finfo = font_asst.get_font_info(key)
                    self.register_font(finfo.name, finfo.font_path,
                                       style=finfo.style, weight=finfo.weight)

        def get_screen_size(self):
            return self.default_session.get_screen_size()

        def get_url(self):
            if not self.use_http_server:
                return None
            base_url = f"http://{self._host}:{self._http_port}/?session={_session_id}&token={_session.token}"
            return base_url

        def mainloop(self, timeout=None):
            super().start()
            super().run()

        def start(self):
            super().start()

        def process_events(self, timeout=0.1):
            super().process_events(timeout=timeout)


# BASIC WIDGETS

def get_args(args):
    """Return the argument tuple appropriate for how we are running."""
    if args is None:
        return tuple([])
    if in_situ_web:
        return args
    # <--- we are running externally, so we need to prepend the default
    #       session
    if len(args) == 0:
        return (_session,)
    return (_session,) + args


class TextEntry(WidgetMixin, PGW.TextEntry):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.TextEntry.__init__(self, *get_args(args), **kwargs)

        # remapping pgwidgets 'activated' to ours
        self.on('activated', self._cb_redirect)
        self._enable_callback('activated')

    def _cb_redirect(self, value):
        # our 'activated' callback has a slightly different callback
        self._make_callback('activated')


class TextEntrySet(WidgetMixin, PGW.TextEntrySet):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        # translate the default "Set" button label (a declared option)
        kwargs.setdefault('label', _tr('Set'))
        PGW.TextEntrySet.__init__(self, *get_args(args), **kwargs)

        # remapping pgwidgets 'activated' to ours
        self.on('activated', self._cb_redirect)
        self._enable_callback('activated')

    def _cb_redirect(self, value):
        # our 'activated' callback has a slightly different callback
        self._make_callback('activated')


class TextArea(WidgetMixin, PGW.TextArea):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.TextArea.__init__(self, *get_args(args), **kwargs)

    def append_text(self, text, autoscroll=True):
        # if text.endswith('\n'):
        #     text = text[:-1]
        super().append_text(text)
        if autoscroll:
            super().set_scroll_position(0.0, 1.0)


class TextSource(WidgetMixin, PGW.TextSource):
    """Ginga wrapper over the pgwidgets TextSource.  The pgwidgets side is
    Python-authoritative (its local text model works before the browser
    connects); this wrapper adds the Ginga TextSource API names that differ
    from pgwidgets (show_line_numbers, append_text, ...) and relays
    callbacks.  Most methods (get_text, insert_text, create_ref, tags,
    find, get_ref_*, ...) are inherited unchanged from PGW.TextSource."""

    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        # ginga uses a boolean (or 'char'/'word'/...) for wrap; pgwidgets
        # wants a mode string ('none'/'word'/'hard').
        if 'wrap' in kwargs:
            kwargs['wrap'] = self._wrap_mode(kwargs['wrap'])
        PGW.TextSource.__init__(self, *get_args(args), **kwargs)

        for name in ['tooltip', 'changed', 'cursor_moved', 'key-press',
                     'line-clicked']:
            self._enable_callback(name)

        # Relay the pgwidgets/model callbacks out through this wrapper.
        self.on('changed', lambda *a: self._make_callback('changed'))
        self.on('cursor_moved',
                lambda *a: self._make_callback('cursor_moved'))
        self.on('line_clicked',
                lambda lineno, *a: self._make_callback('line-clicked', lineno))
        self.on('tooltip', self._tt_cb)

    # -- Ginga TextSource API names that differ from pgwidgets --
    def append_text(self, text, autoscroll=True, tags=None):
        end = self.get_ref_end()
        self.insert_text(end, text, tags=tags)
        self.remove_ref(end)
        if autoscroll:
            self.scroll_to_end()

    @staticmethod
    def _wrap_mode(kind):
        """Translate ginga's wrap value to a pgwidgets wrap mode."""
        if kind in ('none', 'word', 'hard'):
            return kind
        # bool (or anything truthy/falsey): False/none -> 'none', else 'word'
        return 'word' if kind else 'none'

    def set_wrap(self, kind):
        return PGW.TextSource.set_wrap(self, self._wrap_mode(kind))

    def set_limit(self, numlines):
        # matches the qt backend, which also does not cap the buffer here
        pass

    def set_font(self, font, size=10):
        family = font if isinstance(font, str) else getattr(font, 'family',
                                                            str(font))
        return PGW.TextSource.set_font(self, family, size)

    def set_scroll_pos(self, pos):
        # pos is a vertical fraction (0..1); pgwidgets takes (h_pct, v_pct)
        try:
            self.set_scroll_position(0.0, float(pos))
        except (TypeError, ValueError):
            pass

    def show_line_numbers(self, tf):
        return self.set_line_numbers(tf)

    def enable_line_icons(self, tf):
        return PGW.TextSource.set_icon_gutter(self, tf)

    def enable_tooltips(self, tf):
        return self.set_tooltips_enabled(tf)

    def _tt_cb(self, line_no, pos_in_line, text, x, y):
        # Mirror the qt backend: give the page a chance to supply a tooltip
        # string by appending to `res`; show it (or hide, when empty).
        res = []
        self._make_callback('tooltip', res, line_no, pos_in_line, text)
        self.show_tooltip(res[0] if len(res) > 0 else '', x, y)

    def set_icon_gutter(self, tf, pad_px=24):
        return PGW.TextSource.set_icon_gutter(self, tf)


class Label(WidgetMixin, PGW.Label):
    def __init__(self, *args, style='normal', **kwargs):
        WidgetMixin.__init__(self)
        # A 'clickable' label is interactive, so pgwidgets emits pointer
        # events ('click'/'dblclick'/...) on it.  Plain labels are not
        # interactive and only have base callbacks.
        interactive = kwargs.pop('interactive', style == 'clickable')
        PGW.Label.__init__(self, *get_args(args), interactive=interactive,
                           **kwargs)

        # Enable these unconditionally (mirroring the qt backend); for a
        # non-interactive label they simply never fire.
        for name in ['activated', 'released']:
            self._enable_callback(name)

        # Bridge pgwidgets pointer events to ours, but only for an
        # interactive label (a non-interactive one has no 'click').
        if interactive:
            self.on('click', self._cb_redirect_activated)
            self.on('pointer-up', self._cb_redirect_released)

    def _cb_redirect_activated(self, *args):
        self._make_callback('activated')

    def _cb_redirect_released(self, *args):
        self._make_callback('released')


class HSeparator(WidgetMixin, PGW.Separator):
    """A thin horizontal rule (like an HTML <hr>)."""
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.Separator.__init__(self, *get_args(('horizontal',)), **kwargs)


class VSeparator(WidgetMixin, PGW.Separator):
    """A thin vertical rule."""
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.Separator.__init__(self, *get_args(('vertical',)), **kwargs)


class Button(WidgetMixin, PGW.Button):

    # Class-level default hover colors.  Each button snapshots these at
    # construction, so bracketing UI creation with set_hover_color(bg, fg)
    # ... set_hover_color(None, None) gives just those buttons a hover
    # highlight.  Mirrors the qt/gtk backends.  See set_hover_color().
    _hover_bg = None
    _hover_fg = None

    def __init__(self, text=None, iconpath=None, iconsize=None, **kwargs):
        WidgetMixin.__init__(self)
        args = () if text is None else (text,)
        PGW.Button.__init__(self, *get_args(args), **kwargs)

        # icon given at construction (like the qt backend)
        if iconpath is not None:
            self.set_icon(iconpath, iconsize=iconsize)

        # apply the class-level hover default in effect when this button
        # was created (pgwidgets Button.set_hover installs the highlight)
        if Button._hover_bg is not None or Button._hover_fg is not None:
            self.set_hover(Button._hover_bg, Button._hover_fg)

    @classmethod
    def set_hover_color(cls, bg=None, fg=None):
        """Set the default hover background/foreground for buttons created
        *after* this call; pass ``(None, None)`` to clear it."""
        cls._hover_bg = bg
        cls._hover_fg = fg

    def set_icon(self, iconpath, iconsize=None):
        wd, ht = 24, 24
        if iconsize is not None:
            wd, ht = iconsize
        icon_uri = None
        if iconpath is not None:
            icon_uri = PgHelp.get_icon(iconpath, size=(wd, ht))
        super().set_icon(icon_uri, (wd, ht))


class ComboBox(WidgetMixin, PGW.ComboBox):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.ComboBox.__init__(self, *get_args(args), **kwargs)

        # remapping pgwidgets 'activated' to ours
        self.on('activated', self._cb_redirect)
        self._enable_callback('activated')

    def _cb_redirect(self, index, value):
        # our 'activated' callback only returns the index of the chosen thing
        self._make_callback('activated', index)


class SpinBox(WidgetMixin, PGW.SpinBox):
    def __init__(self, *args, dtype=int, **kwargs):
        WidgetMixin.__init__(self)
        _dtype = 'float' if dtype is float else 'int'
        PGW.SpinBox.__init__(self, *get_args(args), dtype=_dtype, **kwargs)

        # remapping 'activated' to 'value-changed'
        self.on('activated', self._cb_redirect)
        self._enable_callback('value-changed')

    def _cb_redirect(self, value):
        self._make_callback('value-changed', value)

    def set_limits(self, minval, maxval, incr_value=1):
        PGW.SpinBox.set_limits(self, minval, maxval, incr_value)


class Slider(WidgetMixin, PGW.Slider):
    def __init__(self, *args, dtype=int, **kwargs):
        WidgetMixin.__init__(self)
        _dtype = 'float' if dtype is float else 'int'
        PGW.Slider.__init__(self, *get_args(args), dtype=_dtype, **kwargs)

        # remapping 'activated' to 'value-changed'
        self.on('activated', self._cb_redirect)
        self._enable_callback('value-changed')

    def _cb_redirect(self, value):
        self._make_callback('value-changed', value)

    def set_limits(self, minval, maxval, incr_value=1):
        PGW.Slider.set_limits(self, minval, maxval, incr_value)


class Dial(WidgetMixin, PGW.Dial):
    def __init__(self, *args, dtype=int, **kwargs):
        WidgetMixin.__init__(self)
        _dtype = 'float' if dtype is float else 'int'
        PGW.Dial.__init__(self, *get_args(args), dtype=_dtype, **kwargs)

        # remapping 'activated' to 'value-changed'
        self.on('activated', self._cb_redirect)
        self._enable_callback('value-changed')

    def _cb_redirect(self, value):
        self._make_callback('value-changed', value)

    def set_limits(self, minval, maxval, incr_value=1):
        PGW.SpinBox.set_limits(self, minval, maxval, incr_value)

    def set_icon(self, iconpath, size=None):
        wd, ht = 24, 24
        if size is not None:
            wd, ht = size
        icon_uri = None
        if iconpath is not None:
            icon_uri = PgHelp.get_icon(iconpath, size=(wd, ht))
        super().set_icon(icon_uri, (wd, ht))


class ScrollBar(WidgetMixin, PGW.ScrollBar):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.ScrollBar.__init__(self, *get_args(args), **kwargs)

    def set_value(self, pct):
        self.set_scroll_percent(pct)

    def get_value(self):
        return self.get_scroll_percent()


class CheckBox(WidgetMixin, PGW.CheckBox):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.CheckBox.__init__(self, *get_args(args), **kwargs)


class ToggleButton(WidgetMixin, PGW.ToggleButton):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.ToggleButton.__init__(self, *get_args(args), **kwargs)


class RadioButton(WidgetMixin, PGW.RadioButton):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.RadioButton.__init__(self, *get_args(args), **kwargs)


class Image(WidgetMixin, PGW.Image):
    def __init__(self, *args, native_image=None, **kwargs):
        WidgetMixin.__init__(self)
        PGW.Image.__init__(self, *get_args(args), **kwargs)

        if native_image is not None:
            self.set_image(native_image)

    def load_file(self, path):
        # Build a ``data:`` URI from the file's bytes and hand it to the
        # browser <img>.  SVG is kept scalable; raster formats keep their
        # native resolution.  (There is no ``to_data_uri`` method -- this
        # used to reference one that never existed.)
        ext = os.path.splitext(path)[1].lower().lstrip('.')
        imgtype = 'svg+xml' if ext == 'svg' else (ext or 'png')
        with open(path, 'rb') as in_f:
            data_uri = PgHelp.get_image_src_from_buffer(in_f.read(),
                                                        imgtype=imgtype)
        self.set_image(data_uri)


class ProgressBar(WidgetMixin, PGW.ProgressBar):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.ProgressBar.__init__(self, *get_args(args), **kwargs)


class StatusBar(WidgetMixin, PGW.StatusBar):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.StatusBar.__init__(self, *get_args(args), **kwargs)

    def clear_message(self):
        super().clear()


# Recognised values for the ``widget`` field of a column
# descriptor — kept in sync with the same constant in the qtw /
# gtk3w / gtk4w wrappers.
_CELL_WIDGETS = ('checkbox', 'combobox', 'progress', 'button')


def _treedata_to_plain(obj):
    """Recursively convert a tree/row data structure into plain JSON-/JS-
    serializable Python types for transport to the browser.

    Ginga plugins commonly build tree rows out of ``Bunch`` (and use
    ``Bunch.caselessDict`` for the tree itself), occasionally stash a
    ``set`` in a row, and frequently store numpy scalars/arrays (e.g. a
    value computed from image data) -- none of which serialize.  Convert
    Bunch/caselessDict -> dict, set -> list, and numpy scalars/0-d arrays
    -> Python scalars (numpy arrays -> list) so the pg backend can ship
    them.  Without the numpy coercion such a value would cross to the
    browser as an opaque object and a TreeView would mistake it for a
    nested child row.  The caller's original structure is left untouched
    (a copy is returned).
    """
    if isinstance(obj, dict):
        return {k: _treedata_to_plain(v) for k, v in obj.items()}
    if isinstance(obj, (Bunch.Bunch, Bunch.caselessDict)):
        return {k: _treedata_to_plain(obj[k]) for k in obj.keys()}
    if isinstance(obj, (set, frozenset)):
        return [_treedata_to_plain(v) for v in obj]
    if isinstance(obj, (list, tuple)):
        return [_treedata_to_plain(v) for v in obj]
    # numpy scalars / 0-d arrays -> Python scalar; larger numpy arrays ->
    # list (duck-typed, mirroring the websocket JsonEncoder, so no hard
    # numpy dependency here).
    item = getattr(obj, 'item', None)
    if callable(item):
        try:
            return item()
        except (TypeError, ValueError):
            pass
    tolist = getattr(obj, 'tolist', None)
    if callable(tolist):
        try:
            return _treedata_to_plain(tolist())
        except (TypeError, ValueError):
            pass
    return obj


def _treedata_to_bunch(obj):
    """Recursively wrap the dicts returned from the browser back into
    ``Bunch`` objects, so the pg TreeView honors the same "Bunch in, Bunch
    out" contract as the qt/gtk backends -- i.e. callers can read results
    with either attribute (``row.NAME``) or item (``row['NAME']``) syntax.
    """
    if isinstance(obj, dict):
        return Bunch.Bunch({k: _treedata_to_bunch(v) for k, v in obj.items()})
    if isinstance(obj, (list, tuple)):
        return [_treedata_to_bunch(v) for v in obj]
    return obj


class TreeView(WidgetMixin, PGW.TreeView):
    def __init__(self, *args, auto_expand=False, sortable=False,
                 selection='single', use_alt_row_color=False,
                 dragable=False):
        WidgetMixin.__init__(self)

        PGW.TreeView.__init__(self, *get_args(args),
                              selection_mode=selection, sortable=sortable,
                              alternate_row_colors=use_alt_row_color)
        self.levels = 1
        self.datakeys = []

        # remapping 'collapsed'
        self.on('collapsed', self._cb_redirect_collapsed)
        self._enable_callback('collapsed')
        self.on('expanded', self._cb_redirect_expanded)
        self._enable_callback('expanded')
        self.on('selected', self._cb_redirect_selected)
        self._enable_callback('selected')
        self.on('activated', self._cb_redirect_activated)
        self._enable_callback('activated')

    def _cb_redirect_collapsed(self, node_vals, path):
        self._make_callback('collapsed', path)

    def _cb_redirect_expanded(self, node_vals, path):
        self._make_callback('expanded', path)

    def _cb_redirect_selected(self, sel_lst):
        subtree = _treedata_to_bunch(super().get_subtree(status='selected'))
        self._make_callback('selected', subtree)

    def _cb_redirect_activated(self, node_vals, path, col_key=None):
        # pgwidgets-js sends ``col_key`` as a 3rd arg on newer
        # builds; older builds send only two.  TreeView's
        # public ``activated`` signature stays ``(widget, subtree)``,
        # so col_key is accepted but not surfaced here.
        subtree = _treedata_to_bunch(super().get_subtree(status=path))
        self._make_callback('activated', subtree)

    def setup_table(self, columns, levels, leaf_key):
        self.levels = levels

        # create the column headers
        col_defs = []
        # columns specifies a mapping
        for col in columns:
            col_def = dict(label=col[0], key=col[1])
            if len(col) > 2:
                col_def['type'] = 'string' if col[2] == 'str' else col[2]
            else:
                # detect an icon column by its data *key* (col[1]), matching
                # the qt/gtk convention -- not the column label (col[0]).
                col_def['type'] = 'icon' if col[1] == 'icon' else 'string'
            col_defs.append(col_def)
        self.datakeys = [col_def['key'] for col_def in col_defs]
        self.leaf_key = leaf_key

        try:
            super().set_columns(col_defs)
        except Exception as e:
            self.logger.error(f"error setting columns: {e}")

    def set_cell_padding(self, px):
        # convenience matching the desktop backends: set both the row
        # (vertical) and column (horizontal) cell spacing.  The
        # set_row_spacing/set_column_spacing primitives are inherited
        # from the underlying pgwidgets TreeView.
        self.set_row_spacing(px)
        self.set_column_spacing(px)

    def set_tree(self, tree_dict):
        super().set_tree(_treedata_to_plain(tree_dict))

    def add_tree(self, tree_dict, expand_new=False):
        super().add_tree(_treedata_to_plain(tree_dict))

    def update_tree(self, tree_dict, expand_new=False):
        super().update_tree(_treedata_to_plain(tree_dict))

    def expand_all(self, tf):
        if tf:
            super().expand_all()
        else:
            super().collapse_all()

    def highlight_path(self, path, onoff, font_color='green'):
        # TODO?
        super().select_path(path, onoff)

    def get_children(self, status='expanded'):
        return _treedata_to_bunch(super().get_subtree(status=status))

    def get_expanded(self):
        return _treedata_to_bunch(super().get_subtree(status='expanded'))

    def get_collapsed(self):
        return _treedata_to_bunch(super().get_subtree(status='collapsed'))

    def get_selected(self):
        return _treedata_to_bunch(super().get_subtree(status='selected'))

    def get_selected_paths(self):
        return super().get_selected()


class TableView(WidgetMixin, PGW.TableView):
    """Flat tabular view, API-compatible with the qtw TableView.

    Thin wrapper over :class:`pgwidgets.PGW.TableView` (which is in
    turn a flat-display subclass of the underlying TreeView) that
    normalises the constructor signature and callback shapes so a
    single piece of caller code works under either pgw or qtw.

    Constructor signature, callbacks, and method names mirror
    ``ginga.qtw.Widgets.TableView`` — see that class's docstring
    for the full API.  Callback signatures emitted here:

    * ``activated(table, row_dict, path, col_key)``
    * ``selected(table, list_of_row_dicts)``
    * ``sorted(table, col_key, ascending)``
    * ``cell_edited(table, path, col_key, old_value, new_value)``
    * ``scrolled(table, h_pct, v_pct)``

    ``path`` is ``[row_index]`` (a length-1 list) for a flat
    table, matching the qtw side.  ``col_key`` on ``activated``
    reflects the clicked cell's column key on a row dblclick, or
    ``None`` when activation came from the Enter key (no specific
    column) — matching qtw/gtk behaviour.
    """

    def __init__(self, *args, columns=None, show_header=True,
                 selection_mode='single', alternate_row_colors=False,
                 show_grid=False, show_row_numbers=False,
                 sortable=False, allow_text_selection=False,
                 dragable=False):
        WidgetMixin.__init__(self)

        # Build the underlying PGW.TableView with the options it
        # understands.  ``dragable`` has no equivalent on the pgw
        # side today; accepted for signature parity and ignored.
        pgw_kwargs = dict(
            show_header=show_header,
            selection_mode=selection_mode,
            alternate_row_colors=alternate_row_colors,
            show_grid=show_grid,
            show_row_numbers=show_row_numbers,
            sortable=sortable,
            allow_text_selection=allow_text_selection,
        )
        if columns is not None:
            pgw_kwargs['columns'] = self._normalise_columns(columns)
        PGW.TableView.__init__(self, *get_args(args), **pgw_kwargs)

        # Stash for callback redirects (path → row_dict lookup).
        self._user_columns = list(pgw_kwargs.get('columns', []) or [])
        # Shadow copy of row data so get_rows / get_row can answer
        # without a round-trip to the JS side (mirrors what the qtw
        # wrapper does by reading from QTreeWidget items).  Kept in
        # sync by the data-mutating method overrides below.
        self._rows = []
        # Parallel mirror of the JS-side row keys, in visible order.
        # After ``set_rows`` these are ``['row0', 'row1', ...]``;
        # ``insert_row`` / ``delete_row`` keep them in lockstep with
        # JS's auto-keying algorithm so ``_to_pgw_path`` /
        # ``_from_pgw_path`` can translate between integer visible
        # positions and JS row keys after arbitrary mutations.
        self._row_keys = []

        # Wire callback redirects so a single handler signature
        # works regardless of widget set.
        self.on('selected', self._cb_redirect_selected)
        self._enable_callback('selected')
        self.on('activated', self._cb_redirect_activated)
        self._enable_callback('activated')
        self.on('sorted', self._cb_redirect_sorted)
        self._enable_callback('sorted')
        self.on('cell_edited', self._cb_redirect_cell_edited)
        self._enable_callback('cell_edited')
        self.on('scrolled', self._cb_redirect_scrolled)
        self._enable_callback('scrolled')
        # Cell-selection + clipboard callbacks (cell modes).
        self.on('cell_selected', self._cb_redirect_cell_selected)
        self._enable_callback('cell_selected')
        # Action-shaped widget cells (currently: button) fire
        # cell_action(table, row_dict, col_key) on click.
        self.on('cell_action', self._cb_redirect_cell_action)
        self._enable_callback('cell_action')
        for name in ('copy', 'cut', 'paste'):
            self.on(name, self._cb_redirect_clipboard, name)
            self._enable_callback(name)

    # ----- column descriptor normalisation -------------------

    @staticmethod
    def _normalise_columns(columns):
        """Accept dicts, tuples, or strings (same forms as the qtw
        wrapper) and return a list of dicts the underlying
        pgwidgets-js TableView understands.

        Optional dict keys: ``halign``, ``editable``, ``widget``
        (one of ``_CELL_WIDGETS`` or None) and widget-specific
        extras: ``choices``, ``min``, ``max``, ``text``.
        """
        out = []
        for i, col in enumerate(columns):
            if isinstance(col, dict):
                key = col.get('key') or col.get('label') or f'col{i}'
                widget = col.get('widget')
                if widget is not None and widget not in _CELL_WIDGETS:
                    raise ValueError(
                        f"unknown column widget {widget!r} "
                        f"(expected one of {_CELL_WIDGETS})")
                d = {
                    'label': col.get('label', key),
                    'key': key,
                    'type': col.get('type', 'string'),
                }
                if 'halign' in col:
                    d['halign'] = col['halign']
                if 'editable' in col:
                    d['editable'] = bool(col['editable'])
                if 'colwidth' in col:
                    d['colwidth'] = col['colwidth']
                if widget is not None:
                    d['widget'] = widget
                    if 'choices' in col:
                        d['choices'] = list(col['choices'])
                    for opt in ('min', 'max', 'text',
                                'enabled_key', 'visible_key'):
                        if opt in col:
                            d[opt] = col[opt]
            elif isinstance(col, (tuple, list)):
                label = col[0]
                key = col[1] if len(col) > 1 else label
                dtype = col[2] if len(col) > 2 else 'string'
                d = {'label': label, 'key': key, 'type': dtype}
            elif isinstance(col, str):
                d = {'label': col, 'key': col, 'type': 'string'}
            else:
                raise ValueError(
                    f"unrecognised column descriptor: {col!r}")
            out.append(d)
        return out

    # ----- column / row API (delegated, with normalisation) --

    def set_columns(self, columns):
        norm = self._normalise_columns(columns)
        self._user_columns = norm
        super().set_columns(norm)

    def append_column(self, column):
        norm, = self._normalise_columns([column])
        self._user_columns.append(norm)
        super().append_column(norm)

    def insert_column(self, idx, column):
        norm, = self._normalise_columns([column])
        self._user_columns.insert(idx, norm)
        super().insert_column(idx, norm)

    def delete_column(self, idx):
        if 0 <= idx < len(self._user_columns):
            del self._user_columns[idx]
        super().delete_column(idx)

    # ----- row data (mutators shadow _rows for get_rows) -----

    def _col_keys(self):
        return [c['key'] for c in self._user_columns]

    def _normalise_row(self, row):
        if isinstance(row, dict):
            return dict(row)
        if isinstance(row, (list, tuple)):
            return dict(zip(self._col_keys(), row))
        raise ValueError(
            f"row must be a dict or sequence, got {type(row).__name__}")

    def _next_row_key(self):
        """Compute the next free ``rowN`` key, matching JS's
        auto-keying algorithm: start at the current row count and
        walk forward past any collisions.  Must be called *before*
        appending the new row to ``_row_keys``."""
        used = set(self._row_keys)
        i = len(self._row_keys)
        while f'row{i}' in used:
            i += 1
        return f'row{i}'

    def set_rows(self, rows):
        self._rows = [self._normalise_row(r) for r in rows]
        self._row_keys = [f'row{i}' for i in range(len(self._rows))]
        super().set_rows(rows)

    def set_data(self, data):
        """Replace all rows.  Accepts the same row forms as :meth:`set_rows`,
        or a numpy array: a structured/record array becomes dict rows (keyed
        by field name), a 2-D array becomes positional rows."""
        from ginga.util import tablehelper
        if tablehelper.is_ndarray(data):
            data = tablehelper.rows_from_ndarray(data)
        self._rows = [self._normalise_row(r) for r in data]
        self._row_keys = [f'row{i}' for i in range(len(self._rows))]
        super().set_data(data)

    def set_table(self, table):
        """Populate the table from an astropy Table or pandas DataFrame:
        derive the columns from the table (name -> key/label, dtype -> type),
        then load its contents as rows."""
        from ginga.util import tablehelper
        self.set_columns(tablehelper.columns_from_table(table))
        self.set_rows(tablehelper.rows_from_table(table))

    def set_cell_padding(self, px):
        # convenience matching the desktop backends: set both the row
        # (vertical) and column (horizontal) cell spacing.  The
        # set_row_spacing/set_column_spacing primitives are inherited
        # from the underlying pgwidgets TableView (a TreeView subclass).
        self.set_row_spacing(px)
        self.set_column_spacing(px)

    def append_row(self, row):
        key = self._next_row_key()
        self._rows.append(self._normalise_row(row))
        self._row_keys.append(key)
        super().append_row(row)

    def insert_row(self, idx, row):
        key = self._next_row_key()
        self._rows.insert(idx, self._normalise_row(row))
        self._row_keys.insert(idx, key)
        super().insert_row(idx, row)

    def delete_row(self, idx):
        if 0 <= idx < len(self._rows):
            del self._rows[idx]
            del self._row_keys[idx]
        super().delete_row(idx)

    def set_cell(self, row, col, value):
        if 0 <= row < len(self._rows):
            if isinstance(col, int):
                if 0 <= col < len(self._user_columns):
                    self._rows[row][self._user_columns[col]['key']] = value
            else:
                self._rows[row][col] = value
        super().set_cell(row, col, value)

    def clear(self):
        self._rows = []
        self._row_keys = []
        super().clear()

    def get_rows(self):
        """Return all rows as a list of dicts.  Mirrors the qtw
        wrapper's method so the same caller code works either way."""
        return [dict(r) for r in self._rows]

    def get_row(self, idx):
        if not (0 <= idx < len(self._rows)):
            raise IndexError(f"row index {idx} out of range")
        return dict(self._rows[idx])

    def set_column_editable(self, col, tf):
        """Mark a column as editable.

        ``col`` may be either an integer user-space column index
        (matching qtw/gtk) or a string column key (matching the
        native pgwidgets-js API).  We accept both so the same
        caller code works under any backend.
        """
        if isinstance(col, int):
            if not (0 <= col < len(self._user_columns)):
                return
            col_key = self._user_columns[col]['key']
        else:
            col_key = col
        # Keep the local _user_columns descriptor in sync so a
        # later set_columns() round-trip preserves the flag.
        for c in self._user_columns:
            if c['key'] == col_key:
                c['editable'] = bool(tf)
                break
        super().set_column_editable(col_key, bool(tf))

    # get_row_count, get_column_count, select_*, sort_by_column,
    # scroll_*, set_show_grid, set_show_row_numbers, set_sortable,
    # set_column_editable, set_column_width,
    # set_optimal_column_widths, set_scroll_position, and
    # get_scroll_position are all inherited from PGW.TableView
    # unchanged.

    # ----- selection conversion ------------------------------

    # ---- path normalisation ---------------------------------
    #
    # pgwidgets-js's TableView.set_data tags each row internally
    # with a string key ``"row0"``, ``"row1"``, ..., and surfaces
    # paths as ``["row0"]`` style lists.  The qtw wrapper, in
    # contrast, uses bare integer row indices (``[0]``, ``[1]``).
    # We convert at the boundary so portable caller code can use
    # integer paths uniformly under either widget set.

    def _to_pgw_path(self, path):
        """Convert a wrapper-style integer-index path (``[2]``) to
        a JS-side key path (``["row5"]``).  Uses our ``_row_keys``
        mirror — naïvely formatting ``'row{N}'`` would be wrong
        after any insert/delete, since JS auto-keys past collisions
        rather than reusing position numbers."""
        # tolerate a bare row index (``2``) in place of a path (``[2]``),
        # matching the qtw wrapper's _resolve_to_row_index
        if isinstance(path, int):
            path = [path]
        out = []
        for i, k in enumerate(path):
            if i == 0 and isinstance(k, int):
                if 0 <= k < len(self._row_keys):
                    out.append(self._row_keys[k])
                else:
                    out.append(f'row{k}')  # best-effort fallback
            elif isinstance(k, int):
                out.append(f'row{k}')
            else:
                out.append(k)
        return out

    def _from_pgw_path(self, path):
        """Inverse of ``_to_pgw_path``: convert ``["row5"]`` to
        ``[2]`` by looking the key up in our mirror.  Falls back
        to ``int(key[3:])`` only when the key isn't tracked — that
        preserves behaviour for paths from a tree we never set
        through ``set_rows`` (rare; mostly defensive)."""
        out = []
        for i, k in enumerate(path or ()):
            if i == 0 and isinstance(k, str):
                if k in self._row_keys:
                    out.append(self._row_keys.index(k))
                    continue
                if k.startswith('row'):
                    try:
                        out.append(int(k[3:]))
                        continue
                    except ValueError:
                        pass
            out.append(k)
        return out

    def get_selected(self):
        """Return selected rows as a list of dicts.

        The underlying ``PGW.TableView.get_selected`` returns a
        list of ``{path, values}`` records; we project to just the
        row dicts for parity with the qtw wrapper.
        """
        raw = super().get_selected()
        out = []
        for entry in raw or []:
            if isinstance(entry, dict) and 'values' in entry:
                out.append(entry['values'])
            else:
                out.append(entry)
        return out

    def get_selected_paths(self):
        """Return rows containing the current selection as a list
        of ``[row_index]`` paths.  Works for both row-mode and
        cell-mode selections — in cell mode the underlying row-
        level selection is empty, so we fall back to deriving row
        indices from the union of selected cells (matching what
        the qtw side returns from ``selectedItems()``)."""
        raw = super().get_selected()
        out = [self._from_pgw_path(entry['path'])
               for entry in (raw or [])
               if isinstance(entry, dict) and 'path' in entry]
        if out:
            return out
        # Cell-mode fallback: derive unique row paths from the
        # cell selection.
        seen = set()
        for c in self.get_selected_cells():
            path = c.get('path')
            if not path:
                continue
            key = tuple(path)
            if key in seen:
                continue
            seen.add(key)
            out.append(list(path))
        out.sort()
        return out

    def select_path(self, path, state=True):
        super().select_path(self._to_pgw_path(path), state)

    def select_paths(self, paths, state=True):
        super().select_paths([self._to_pgw_path(p) for p in paths],
                             state)

    def scroll_to_path(self, path):
        super().scroll_to_path(self._to_pgw_path(path))

    # ----- callback redirects --------------------------------

    def _cb_redirect_selected(self, sel_lst):
        # PGW emits the raw [{path, values}, ...] list; the
        # wrapper's get_selected projects to row dicts.
        self._make_callback('selected', self.get_selected())

    def _cb_redirect_activated(self, values, path, col_key=None):
        # PGW signature: (values_dict, path, col_key).  Re-emit as
        # (row_dict, path, col_key) so it matches qtw/gtk, with path
        # normalised to integer row indices.  ``col_key`` defaults
        # to None for older pgwidgets-js builds that don't yet send
        # the clicked column, and is also None when activation came
        # from the Enter key rather than a cell dblclick.
        self._make_callback('activated', values,
                            self._from_pgw_path(path), col_key)

    def _cb_redirect_sorted(self, col_key, ascending):
        self._make_callback('sorted', col_key, ascending)

    def _cb_redirect_cell_edited(self, path, col_key,
                                 old_value, new_value):
        self._make_callback('cell_edited',
                            self._from_pgw_path(path), col_key,
                            old_value, new_value)

    def _cb_redirect_scrolled(self, h_pct, v_pct):
        self._make_callback('scrolled', h_pct, v_pct)

    def _cb_redirect_cell_action(self, path, col_key):
        """JS fires ``cell_action(path, col_key)`` when the user
        clicks a button-shaped widget cell.  Resolve ``path`` to
        the row dict and re-emit as ``cell_action(table, row_dict,
        col_key)`` to match the qtw signature."""
        idx_path = self._from_pgw_path(path)
        idx = idx_path[0] if idx_path else None
        row = (dict(self._rows[idx])
               if idx is not None and 0 <= idx < len(self._rows)
               else None)
        self._make_callback('cell_action', row, col_key)

    def _cb_redirect_cell_selected(self, cells):
        # JS sends ``[{path, col_key, value}, ...]`` with pgw-style
        # paths.  Normalise paths to integer row indices to match
        # the qtw side.
        out = []
        for c in (cells or []):
            if isinstance(c, dict):
                d = dict(c)
                if 'path' in d:
                    d['path'] = self._from_pgw_path(d['path'])
                out.append(d)
        self._make_callback('cell_selected', out)

    def _cb_redirect_clipboard(self, tsv, name):
        # ``copy``/``cut``/``paste`` carry the TSV (or pasted text)
        # exactly as the JS produced it.  Re-emit unchanged.
        self._make_callback(name, tsv)

    # ----- cell-selection API (matches the qtw wrapper) -------

    def get_selected_cells(self):
        """Return ``[{path, col_key, value}, ...]`` for the current
        cell selection, with paths normalised to integer row
        indices."""
        raw = super().get_selected_cells()
        out = []
        for c in (raw or []):
            if isinstance(c, dict):
                d = dict(c)
                if 'path' in d:
                    d['path'] = self._from_pgw_path(d['path'])
                out.append(d)
        return out

    def select_cell(self, path, col_key, state=True):
        super().select_cell(self._to_pgw_path(path), col_key, bool(state))

    def select_cells(self, cells, state=True):
        converted = []
        for c in (cells or []):
            converted.append({
                'path': self._to_pgw_path(c.get('path', [])),
                'col_key': c.get('col_key'),
            })
        super().select_cells(converted, bool(state))

    def clear_cell_selection(self):
        super().clear_cell_selection()

    # ``copy_selection`` / ``cut_selection`` / ``paste_selection``
    # are exposed by pgwidgets-js as auto-dispatched action methods
    # (see ``pgwidgets_js/defs.py``).  pgwidgets-python's class
    # factory generates the matching Python methods automatically,
    # so the inherited ones forward the call into the browser and
    # fire the ``copy`` / ``cut`` / ``paste`` callbacks once the
    # JS-side clipboard work completes.  No wrapper override
    # needed.

    # ----- per-cell / row / column / table colour overrides --
    #
    # ``set_column_color`` / ``set_table_color`` / ``clear_column_color``
    # / ``clear_all_colors`` don't need path translation and so are
    # auto-generated from defs.py without further help here.  The
    # path-taking methods need our integer-index ↔ pgw-row-key
    # boundary conversion.

    def set_cell_color(self, path, col_key, fg=None, bg=None, bold=None):
        super().set_cell_color(self._to_pgw_path(path), col_key,
                               fg=fg, bg=bg, bold=bold)

    def set_row_color(self, path, fg=None, bg=None, bold=None):
        super().set_row_color(self._to_pgw_path(path),
                              fg=fg, bg=bg, bold=bold)

    def clear_cell_color(self, path, col_key):
        super().clear_cell_color(self._to_pgw_path(path), col_key)

    def clear_row_color(self, path):
        super().clear_row_color(self._to_pgw_path(path))


class Canvas(WidgetMixin, PGW.Canvas):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.Canvas.__init__(self, *get_args(args), **kwargs)


class Box(ContainerWidgetMixin, PGW.Box):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.Box.__init__(self, *get_args(args), **kwargs)


class HBox(ContainerWidgetMixin, PGW.HBox):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.HBox.__init__(self, *get_args(args), **kwargs)


class VBox(ContainerWidgetMixin, PGW.VBox):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.VBox.__init__(self, *get_args(args), **kwargs)


class ButtonBox(ContainerWidgetMixin, PGW.ButtonBox):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.ButtonBox.__init__(self, *get_args(args), **kwargs)


class FixedLayout(ContainerWidgetMixin, PGW.FixedLayout):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.FixedLayout.__init__(self, *get_args(args), **kwargs)


class Frame(ContainerWidgetMixin, PGW.Frame):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.Frame.__init__(self, *get_args(args), **kwargs)


class Expander(ContainerWidgetMixin, PGW.Expander):
    def __init__(self, *args, notoggle=False, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.Expander.__init__(self, *get_args(args),
                              collapsible=not notoggle, **kwargs)

    def expand(self, tf):
        self.set_collapsed(not tf)


class TabWidget(ContainerWidgetMixin, PGW.TabWidget):
    def __init__(self, *args, tabpos='top', reorderable=False,
                 detachable=False, closable=False, group=0):
        ContainerWidgetMixin.__init__(self)
        PGW.TabWidget.__init__(self, *get_args(args), closable=closable,
                               reorderable=reorderable, tab_position=tabpos)

        # remapping 'page_switch'
        self.on('page-switch', self._cb_redirect)
        self._enable_callback('page-switch')

    def add_widget(self, child, title=''):
        # pgwidgets' add_widget takes (child, options); pack the tab
        # title into the options object it expects
        return super().add_widget(child, dict(title=title))

    def _cb_redirect(self, child, index):
        self._make_callback('page-switch', child)


class StackWidget(ContainerWidgetMixin, PGW.StackWidget):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.StackWidget.__init__(self, *get_args(args), **kwargs)

        # remapping 'page_switch'
        self.on('page-switch', self._cb_redirect)
        self._enable_callback('page-switch')

    def add_widget(self, child, title=''):
        # pgwidgets' add_widget takes (child, options); pack the tab
        # title into the options object it expects
        return super().add_widget(child, dict(title=title))

    def _cb_redirect(self, child, index):
        self._make_callback('page-switch', child)


class MDIWidget(ContainerWidgetMixin, PGW.MDIWidget):
    def __init__(self, *args, tabpos='top', mode='mdi', **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.MDIWidget.__init__(self, *get_args(args), **kwargs)

        self.true_mdi = True
        self.icon_data_uri = PgHelp.get_icon(app_icon_path, size=(16, 16))

    def add_widget(self, child, title=''):
        # pgwidgets' add_widget takes (child, options); pack the title
        # and window icon into the options object it expects
        return super().add_widget(child, dict(title=title,
                                              icon_url=self.icon_data_uri))

    def tile_panes(self):
        super().tile_windows()

    def cascade_panes(self):
        super().cascade_windows()


# class MDISubWindow(ContainerWidgetMixin, PGW.MDISubWindow):
#     def __init__(self, *args, **kwargs):
#         ContainerWidgetMixin.__init__(self)
#         PGW.MDISubWindow.__init__(self, *get_args(args), **kwargs)

#     def move(self, x, y):
#         super().set_position(x, y)


class ScrollArea(ContainerWidgetMixin, PGW.ScrollArea):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.ScrollArea.__init__(self, *get_args(args), **kwargs)

    def scroll_to_end(self, vertical=True, horizontal=False):
        h_pct, v_pct = self.get_scroll_position()
        if vertical:
            v_pct = 1.0
        if horizontal:
            h_pct = 1.0
        self.set_scroll_position(h_pct, v_pct)


class AbstractScrollArea(ContainerWidgetMixin, PGW.AbstractScrollArea):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.AbstractScrollArea.__init__(self, *get_args(args), **kwargs)


class Splitter(ContainerWidgetMixin, PGW.Splitter):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.Splitter.__init__(self, *get_args(args), **kwargs)


class GridBox(ContainerWidgetMixin, PGW.GridBox):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.GridBox.__init__(self, *get_args(args), **kwargs)

    def resize_grid(self, rows, columns):
        pass

    def add_widget(self, child, row, col, stretch=0):
        super().add_widget(child, row, col)


class ToolbarAction(WidgetMixin, PGW.ToolBarAction):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.ToolBarAction.__init__(self, *get_args(args), **kwargs)


class Toolbar(ContainerWidgetMixin, PGW.ToolBar):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.ToolBar.__init__(self, *get_args(args), **kwargs)

    def add_menu(self, text, menu=None, mtype='tool'):
        if menu is None:
            menu = Menu()
        if mtype == 'mbar':
            child = self.add_action(text, menu=menu)
        else:
            child = Button(text)
            self.add_widget(child)
            child.add_callback('activated', self._popup_menu, menu)

        return menu

    def _popup_menu(self, child, menu):
        x, y = child.get_position()
        menu.popup(x + 20, y + 20)

    def add_action(self, text, toggle=False, iconpath=None, iconsize=None,
                   menu=None):
        icon_uri = None
        if iconpath is not None:
            icon_uri = PgHelp.get_icon(iconpath, size=iconsize)

        # pgwidgets' add_action takes a single ``options`` object
        options = dict(text=text, icon_url=icon_uri, toggle=toggle)
        if menu is not None:
            options['menu'] = menu
        return super().add_action(options)


class MenuAction(WidgetMixin, PGW.MenuAction):
    def __init__(self, *args, **kwargs):
        WidgetMixin.__init__(self)
        PGW.MenuAction.__init__(self, *get_args(args), **kwargs)


class Menu(ContainerWidgetMixin, PGW.Menu):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.Menu.__init__(self, *get_args(args), **kwargs)

    def add_name(self, name, checkable=False):
        child = MenuAction(text=name, checkable=checkable)
        self.add_widget(child)
        return child

    def add_menu(self, name, menu=None):
        if menu is None:
            menu = Menu()
        super().add_menu(name, menu)
        return menu

    def get_menu(self, name):
        menu = super().get_menu(name)
        if menu is None:
            raise KeyError(name)
        return menu

    def popup(self, *args):
        if len(args) == 0:
            x, y = 0, 0
        elif len(args) == 1:
            widget = args[0]
            x, y = widget.get_position()
            # offset a bit from popup widget
            x, y = x + 20, y + 20
        elif len(args) == 2:
            x, y = args
        super().popup(x, y)


class Menubar(ContainerWidgetMixin, PGW.MenuBar):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.MenuBar.__init__(self, *get_args(args), **kwargs)

    def add_name(self, name):
        child = Menu()
        self.add_menu(child, name)
        return child

    def get_menu(self, name):
        menu = super().get_menu(name)
        if menu is None:
            raise KeyError(name)
        return menu


class Page(ContainerWidgetMixin, PGW.Page):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.Page.__init__(self, *get_args(args), **kwargs)


class TopLevel(ContainerWidgetMixin, PGW.TopLevel):
    def __init__(self, *args, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.TopLevel.__init__(self, *get_args(args), **kwargs)


class Dialog(ContainerWidgetMixin, PGW.Dialog):
    def __init__(self, *args, parent=None, flags=None, **kwargs):
        # `parent` and `flags` are accepted for API parity with the qt/gtk
        # backends but are not needed here: a browser Dialog is a DOM overlay
        # that already renders on top of the page content, so there is no
        # window stacking to manage (cf. DIALOG_FLAGS_ONTOP).
        ContainerWidgetMixin.__init__(self)
        PGW.Dialog.__init__(self, *get_args(args), **kwargs)

    def get_content_area(self):
        # NOTE: this is a bit of a hack--get_content_area is not directly
        # supported via the Python interface. But many of the same methods
        # will work on the Dialog itself
        return self

    def delete(self):
        super().destroy()


class ColorDialog(ContainerWidgetMixin, PGW.ColorDialog):
    def __init__(self, *args, parent=None, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.ColorDialog.__init__(self, *get_args(args), **kwargs)


class FileDialog(ContainerWidgetMixin, FileBrowser):
    """File browser for accessing files locally."""
    def __init__(self, *args, title='', parent=None, **kwargs):
        ContainerWidgetMixin.__init__(self)
        FileBrowser.__init__(self, *get_args(args), title=title)

    def _close_and_fire(self, result):
        # The pgwidgets FileBrowser fires 'activated' with a single path
        # string (mode='file'/'directory') or a list (mode='files').  The
        # Ginga contract (matching the qt/gtk backends) is always a list of
        # paths, so callers can index ``filepaths[0]``.  Normalize here.
        if not isinstance(result, (list, tuple)):
            result = [result]
        super()._close_and_fire(result)


class BrowserFileDialog(ContainerWidgetMixin, PGW.FileDialog):
    """File browser for uploading files to a remote instance."""
    def __init__(self, *args, parent=None, **kwargs):
        ContainerWidgetMixin.__init__(self)
        PGW.FileDialog.__init__(self, *get_args(args))

    def popup(self, *args):
        self.show()


class MessageDialog(Dialog):

    icon_dct = dict()

    @classmethod
    def set_category_icon(cls, category, iconpath, size=(64, 64)):
        cls.icon_dct[category] = iconpath

    def __init__(self, title='', flags=None, buttons=[("Dismiss", 0)],
                 parent=None, modal=False, autoclose=False):
        Dialog.__init__(self, title=title, flags=flags, buttons=buttons,
                        parent=parent, modal=modal, autoclose=autoclose)

        # initialize default icons for certain categories
        if 'warning' not in MessageDialog.icon_dct:
            for category, iconfile in [('warning', "warning.svg"),
                                       #('critical', "critical.svg"),
                                       #('denied', "denied.svg"),
                                       ('error', "error.svg"),
                                       ('info', "information.svg"),
                                       ('question', "question.svg")]:
                iconpath = os.path.join(icondir, iconfile)
                MessageDialog.set_category_icon(category, iconpath)

        vbox = self.get_content_area()
        vbox.set_margins(4, 4, 4, 4)

    def set_message(self, category, text, title=None):
        if title is not None:
            self.set_title(title)
        vbox = self.get_content_area()
        vbox.remove_all()
        if category in self.icon_dct:
            hbox = HBox()
            hbox.set_border_width(4)
            hbox.add_widget(Label(""), stretch=1)
            img = Image()
            img.load_file(self.icon_dct[category])
            hbox.add_widget(img, stretch=0)
            hbox.add_widget(Label(""), stretch=1)
            vbox.add_widget(hbox, stretch=1)

        tw = Label(text)
        vbox.add_widget(tw, stretch=1)
        vbox.add_widget(tw)


class Timer(CallbackMixin, PGW.Timer):
    def __init__(self, *args, **kwargs):
        CallbackMixin.__init__(self)
        if in_situ_web:
            PGW.Timer.__init__(self, **kwargs)
        else:
            PGW.Timer.__init__(self, *get_args(args), **kwargs)

        # for storing random data in a timer
        self.data = Bunch.Bunch()

        # remapping 'expired'
        for name in ['expired', 'cancelled']:
            self.on(name, self._cb_redirect, name)
            self._enable_callback(name)

    def _cb_redirect(self, value, name):
        self._make_callback(name)

    def set(self, val):
        self.start(val)

    def clear(self):
        self.cancel()


# MODULE FUNCTIONS

def name_mangle(name, pfx=''):
    newname = []
    for c in name.lower():
        if not (c.isalpha() or c.isdigit() or (c == '_')):
            newname.append('_')
        else:
            newname.append(c)
    return pfx + ''.join(newname)


def make_widget(title, wtype):
    if wtype == 'label':
        w = Label(title, halign='right')
    elif wtype == 'llabel':
        w = Label(title, halign='left')
    elif wtype in ('textentry', 'entry'):
        w = TextEntry()
        w.set_length(12)
    elif wtype in ('textentryset', 'entryset'):
        w = TextEntrySet()
        w.set_length(12)
    elif wtype == 'combobox':
        w = ComboBox()
        w.set_length(12)
    elif wtype == 'comboboxedit':
        w = ComboBox(editable=True)
    elif wtype in ('spinbox', 'spinbutton'):
        w = SpinBox(dtype=int)
    elif wtype == 'spinfloat':
        w = SpinBox(dtype=float)
    elif wtype == 'vbox':
        w = VBox()
    elif wtype == 'hbox':
        w = HBox()
    elif wtype in ('hslider', 'hscale'):
        w = Slider(orientation='horizontal')
    elif wtype in ('vslider', 'vscale'):
        w = Slider(orientation='vertical')
    elif wtype in ('checkbox', 'checkbutton'):
        w = CheckBox(title)
    elif wtype == 'radiobutton':
        w = RadioButton(title)
    elif wtype == 'togglebutton':
        w = ToggleButton(title)
    elif wtype == 'button':
        w = Button(title)
    elif wtype == 'spacer':
        w = Label('')
    elif wtype == 'textarea':
        w = TextArea(editable=True)
    elif wtype == 'toolbar':
        w = Toolbar()
    elif wtype == 'progress':
        w = ProgressBar()
    elif wtype == 'menubar':
        w = Menubar()
    elif wtype == 'dial':
        w = Dial()
    else:
        raise ValueError("Bad wtype=%s" % wtype)
    return w


def hadjust(w, orientation):
    """Ostensibly, a function to reduce the vertical footprint of a widget
    that is normally used in a vertical stack (usually a Splitter), when it
    is instead used in a horizontal orientation.
    """
    return w


def build_info(captions, orientation='vertical'):
    numrows = len(captions)
    numcols = reduce(lambda acc, tup: max(acc, len(tup)), captions, 0)
    if (numcols % 2) != 0:
        raise ValueError("Column spec is not an even number")
    numcols = int(numcols // 2)

    table = GridBox(rows=numrows, columns=numcols)

    wb = Bunch.Bunch()
    row = 0
    for tup in captions:
        col = 0
        while col < numcols:
            idx = col * 2
            if idx < len(tup):
                title, wtype = tup[idx:idx + 2]
                title, disp = translate_caption(title, wtype)
                if not title.endswith(':'):
                    name = name_mangle(title)
                else:
                    name = name_mangle('lbl_' + title[:-1])
                w = make_widget(disp, wtype)
                table.add_widget(w, row, col)
                wb[name] = w
            col += 1
        row += 1

    w = hadjust(table, orientation=orientation)

    return w, wb


def wrap(native_widget):
    return native_widget


# END



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.txt
#
# Note that not all possible configuration values are present in this file.
#
# All configuration values have a default. Some values are defined in
# the global Astropy configuration which is loaded here before anything else.
# See astropy.sphinx.conf for which values are set there.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('..'))
# IMPORTANT: the above commented section was generated by sphinx-quickstart, but
# is *NOT* appropriate for astropy or Astropy affiliated packages. It is left
# commented out with this explanation to make it clear why this should not be
# done. If the sys.path entry above is added, when the astropy.sphinx.conf
# import occurs, it will import the *source* version of astropy instead of the
# version installed (if invoked as "make html" or directly with sphinx).
# Thus, any C-extensions that are needed to build the documentation will *not*
# be accessible, and the documentation will not build correctly.

import datetime
import os
import sys

from ginga import __version__

# Load all of the global Astropy configuration
try:
    from sphinx_astropy.conf.v1 import *  # noqa
except ImportError:
    print('ERROR: the documentation requires the sphinx-astropy package '
          'to be installed')
    sys.exit(1)

# Get configuration information from setup.cfg
from configparser import ConfigParser
conf = ConfigParser()
conf.read([os.path.join(os.path.dirname(__file__), '..', 'setup.cfg')])
setup_cfg = dict(conf.items('metadata'))

# -- General configuration ----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.2'

# To perform a Sphinx version check that needs to be more specific than
# major.minor, call `check_sphinx_version("x.y.z")` here.
# check_sphinx_version("1.2.1")

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns.append('_templates')  # noqa

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog += """
"""

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['ginga.']

# -- Project information ------------------------------------------------------

# This does not *have* to match the package name, but typically does
project = setup_cfg['name']
author = setup_cfg['author']
copyright = '{0}, {1}'.format(datetime.datetime.now().year, author)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# The full version, including alpha/beta/rc tags.
release = __version__
# The short X.Y version.
version = '.'.join(release.split('.')[:2])

# -- Options for HTML output ---------------------------------------------------

# A NOTE ON HTML THEMES
# The global astropy configuration uses a custom theme, 'bootstrap-astropy',
# which is installed along with astropy. A different theme can be used or
# the options for this theme can be modified by overriding some of the
# variables set in the global configuration. The variables set in the
# global configuration are listed below, commented out.

# Add any paths that contain custom themes here, relative to this directory.
# To use a different custom theme, add the directory containing the theme.
#html_theme_path = []

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes. To override the custom theme, set this to the
# name of a builtin theme or the name of a custom theme in html_theme_path.
html_theme = 'sphinx_rtd_theme'

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/ginga-128x128.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = '_static/ginga.ico'

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = ''

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = '{0} v{1}'.format(project, release)

# Output file base name for HTML help builder.
htmlhelp_basename = project + 'doc'


# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [('index', project + '.tex', project + u' Documentation',
                    author, 'manual')]

# this is for working around a bug in the RTD doc generation for PDF
latex_elements = {
    # Additional stuff for the LaTeX preamble.
    'preamble': "".join((
        r'\DeclareUnicodeCharacter{9280}{N}',
        r'\DeclareUnicodeCharacter{6CB3}{N}',
        r'\DeclareUnicodeCharacter{304E}{N}',
        r'\DeclareUnicodeCharacter{3093}{N}',
        r'\DeclareUnicodeCharacter{304C}{N}',
    )),
}

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [('index', project.lower(), project + u' Documentation',
              [author], 1)]



#
# ipg.py -- Module for simple FITS viewer in an HTML5 canvas web browser.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
This module makes it easy to open a Ginga viewer in a web browser (or
embedded in a Jupyter notebook cell).  All the rendering is done on the
server side and the browser only acts as a display front end, connected
over a WebSocket.  Using this you could create an analysis-type
environment on a server and view it via a browser.

It is built on the ``pgwidgets`` backend (``ginga.web.pgw``), which runs
its WebSocket and HTTP servers in their own background threads.  Because
of that, it coexists cleanly with the Jupyter kernel's own event loop --
there is no need to surrender the kernel's IOLoop.  For browser-initiated
viewer interactions (cursor readout, drawing, pan/zoom) the server pumps
queued callbacks onto the running kernel asyncio loop, so they execute on
the kernel thread, in step with the code you run in cells.

Typical notebook usage::

    from ginga.web.pgw import ipg
    server = ipg.make_server(host='localhost', port=9914)
    server.start(no_ioloop=True)
    v1 = server.get_viewer('v1')
    v1.open()              # open in a new browser tab/window
    v1.embed(height=650)   # ...or embed in the notebook cell
    v1.load('/path/to/image.fits')

You will need a reasonably modern web browser with HTML5 canvas support.
"""

import os
import logging
import threading
import asyncio

from ginga import AstroImage
from ginga.canvas.CanvasObject import get_canvas_types
from ginga.misc import log, Bunch
from ginga.Bindings import ImageViewBindings
from ginga.misc.Settings import SettingGroup
from ginga.util.paths import ginga_home
from ginga.util import loader

from ginga.web.pgw import Widgets, Viewers


class BasicCanvasView(Viewers.CanvasView):

    def build_gui(self, container):
        """
        This is responsible for building the viewer's UI.  It should
        place the UI in `container`.  Override this to make a custom
        UI.
        """
        vbox = Widgets.VBox()
        vbox.set_border_width(0)
        # fill the window in both directions
        vbox.set_expanding(True, True)

        w = Viewers.GingaViewerWidget(viewer=self)
        vbox.add_widget(w, stretch=1)

        container.set_widget(vbox)

    def embed(self, width=600, height=650):
        """
        Embed a viewer into a Jupyter notebook.
        """
        from IPython.display import IFrame
        return IFrame(self.url, width, height)

    def open(self, new=1):
        """
        Open this viewer in a new browser window or tab.
        (requires `webbrowser` module)
        """
        import webbrowser
        webbrowser.open(self.url, new=new)

    def show(self, fmt=None):
        """
        Capture the window of a viewer.
        """
        # force any delayed redraws
        # TODO: this really needs to be addressed in get_rgb_image_as_bytes()
        # of the various superclasses, as it affects other backends as well
        self.redraw_now()

        from IPython.display import Image

        if fmt is None:
            # what format are we using for the HTML5 canvas--use that
            settings = self.get_settings()
            fmt = settings.get('html5_canvas_format', 'png')

        return Image(data=bytes(self.get_rgb_image_as_bytes(format=fmt)),
                     format=fmt, embed=True)

    def load_fits(self, filepath):
        """
        Load a FITS file into the viewer.
        """
        image = loader.load_data(filepath, logger=self.logger)
        self.set_image(image)

    load = load_fits

    def load_hdu(self, hdu):
        """
        Load an HDU into the viewer.
        """
        image = AstroImage.AstroImage(logger=self.logger)
        image.load_hdu(hdu)

        self.set_image(image)

    def load_data(self, data_np):
        """
        Load raw numpy data into the viewer.
        """
        image = AstroImage.AstroImage(logger=self.logger)
        image.set_data(data_np)

        self.set_image(image)

    def add_canvas(self, tag=None):
        # add a canvas to the view
        my_canvas = self.get_canvas()
        DrawingCanvas = my_canvas.get_draw_class('drawingcanvas')
        canvas = DrawingCanvas()
        # enable drawing on the canvas
        canvas.enable_draw(True)
        canvas.enable_edit(True)
        canvas.set_drawtype(None)
        canvas.ui_set_active(True)
        canvas.set_surface(self)
        canvas.register_for_cursor_drawing(self)
        # add the canvas to the view.
        my_canvas.add(canvas, tag=tag)
        canvas.set_draw_mode(None)

        return canvas

    def set_html5_canvas_format(self, fmt):
        """
        Sets the format used for rendering to the HTML5 canvas.
        'png' offers greater clarity, especially for small text, but
        does not have as good of performance as 'jpeg'.
        """
        fmt = fmt.lower()
        if fmt not in ('jpeg', 'png'):
            raise ValueError("Format must be one of {jpeg|png} not '%s'" % (
                fmt))

        settings = self.get_settings()
        settings.set(html5_canvas_format=fmt)

    def get_html5_canvas_format(self):
        settings = self.get_settings()
        return settings.get('html5_canvas_format')


class EnhancedCanvasView(BasicCanvasView):
    """
    Like BasicCanvasView, but includes a readout widget for when the
    cursor is moved over the canvas to display the coordinates.
    """

    def build_gui(self, container):
        """
        This is responsible for building the viewer's UI.  It should
        place the UI in `container`.
        """
        vbox = Widgets.VBox()
        vbox.set_border_width(2)
        vbox.set_spacing(1)
        # fill the window in both directions
        vbox.set_expanding(True, True)

        w = Viewers.GingaViewerWidget(viewer=self)
        vbox.add_widget(w, stretch=1)

        # set up to capture cursor movement for reading out coordinates

        # coordinates reported in base 1 or 0?
        self.pixel_base = 1.0

        self.readout = Widgets.Label("")
        vbox.add_widget(self.readout, stretch=0)

        #self.set_callback('none-move', self.motion_cb)
        self.set_callback('cursor-changed', self.motion_cb)

        container.set_widget(vbox)

    def motion_cb(self, viewer, button, data_x, data_y):

        # Get the value under the data coordinates
        try:
            # We report the value across the pixel, even though the coords
            # change halfway across the pixel
            value = viewer.get_data(int(data_x + 0.5), int(data_y + 0.5))

        except Exception:
            value = None

        pb = self.pixel_base
        fits_x, fits_y = data_x + pb, data_y + pb

        # Calculate WCS RA
        try:
            # NOTE: image function operates on DATA space coords
            image = viewer.get_image()
            if image is None:
                # No image loaded
                return
            ra_txt, dec_txt = image.pixtoradec(fits_x, fits_y,
                                               format='str', coords='fits')
        except Exception as e:
            self.logger.warning("Bad coordinate conversion: %s" % (
                str(e)))
            ra_txt = 'BAD WCS'
            dec_txt = 'BAD WCS'

        text = "RA: %s  DEC: %s  X: %.2f  Y: %.2f  Value: %s" % (
            ra_txt, dec_txt, fits_x, fits_y, value)
        self.readout.set_text(text)

    def set_readout_text(self, text):
        self.readout.set_text(text)


class ViewerFactory:
    """
    This is a factory class that churns out web viewers for a web
    application.

    The most important method of interest is get_viewer().

    Each viewer is built in its own pgwidgets *session* so that it has an
    independent URL that can be opened in a browser tab or embedded in a
    notebook cell without interfering with any other viewer.  The first
    viewer reuses the application's default session.
    """

    def __init__(self, logger, app):
        """
        Parameters
        ----------
        logger : python compatible logger
            a logging-module compatible logger object
        app : ginga pgw web application object
        """
        self.logger = logger
        self.app = app
        self.dc = get_canvas_types()
        # dict of viewers
        self.viewers = {}
        # whether the application's default session has been claimed yet
        self._used_default = False

    def _get_session(self):
        # The first viewer reuses the application's default session; each
        # additional viewer gets its own session so it has an isolated
        # URL (a browser tab or embedded iframe shows only that viewer).
        if not self._used_default:
            self._used_default = True
            return self.app.default_session
        return self.app.create_session()

    def _session_url(self, session):
        return "http://%s:%d/?session=%s&token=%s" % (
            self.app._host, self.app._http_port, session.id, session.token)

    def make_viewer(self, v_id, viewer_class=None, width=512, height=512):
        if viewer_class is None:
            viewer_class = EnhancedCanvasView

        session = self._get_session()
        url = self._session_url(session)

        # Build this viewer's widgets (window, canvas surface, timers) in
        # its own session.  The pgwidgets widgets pick up the active
        # session from the module-level ``Widgets._session`` at creation
        # time, so swap it in for the duration of the build and restore
        # the default afterwards.
        prev_session = Widgets._session
        Widgets._session = session
        try:
            # create the top level window in this session
            window = self.app.make_window("Viewer %s" % v_id)

            # load binding preferences if available
            cfgfile = os.path.join(ginga_home, "ipg_bindings.cfg")
            bindprefs = SettingGroup(name='bindings', logger=self.logger,
                                     preffile=cfgfile)
            bindprefs.load(onError='silent')

            bd = ImageViewBindings(self.logger, settings=bindprefs)

            fi = viewer_class(self.logger, bindings=bd)
            fi.url = url
            # set up some reasonable defaults--user can change these later
            # if desired
            fi.set_autocut_params('zscale')
            fi.enable_autocuts('on')
            fi.enable_autozoom('on')
            fi.set_bg(0.2, 0.2, 0.2)
            fi.ui_set_active(True)
            fi.ipg_parent = self

            # enable most key/mouse operations
            bd = fi.get_bindings()
            bd.enable_all(True)

            # set up a non-private canvas for drawing
            canvas = self.dc.DrawingCanvas()
            canvas.set_surface(fi)
            # add canvas to view
            private_canvas = fi.get_canvas()
            private_canvas.add(canvas)
            canvas.ui_set_active(True)
            fi.set_canvas(canvas)

            fi.set_desired_size(width, height)
            # force allocation of a surface--may be resized later
            fi.configure_surface(width, height)

            # add little mode indicator that shows modal states in
            # the corner
            fi.show_mode_indicator(True)

            # Have the viewer build its UI into the container
            fi.build_gui(window)

            window.resize(width, height)
            window.show()

            # Attaching the viewer widget above resizes the offscreen
            # surface to the (as yet unknown) browser canvas size.  Re-
            # establish the desired size so that ``show()`` produces a
            # correctly-sized snapshot even before a browser has connected
            # to report its real canvas dimensions (a connected browser
            # will reconfigure this on resize).
            fi.configure_surface(width, height)

        finally:
            Widgets._session = prev_session

        v_info = Bunch.Bunch(url=url, viewer=fi, top=window, session=session)
        return v_info

    def get_viewer(self, v_id, viewer_class=None, width=512, height=512,
                   force_new=False):
        """
        Get an existing viewer by viewer id.  If the viewer does not yet
        exist, make a new one.
        """
        if not force_new:
            try:
                return self.viewers[v_id]
            except KeyError:
                pass

        # We get back a record with information about the viewer
        v_info = self.make_viewer(v_id, viewer_class=viewer_class,
                                  width=width, height=height)

        # Save it under this viewer id
        self.viewers[v_id] = v_info
        return v_info

    def delete_viewer(self, v_id):
        del self.viewers[v_id]

    def delete_all_viewers(self):
        self.viewers = {}


class WebServer:
    """
    Manages the lifetime of the pgwidgets web application that hosts the
    viewers.

    The pgwidgets ``Application`` runs its WebSocket and HTTP servers in
    their own background threads, so :meth:`start` does not block and does
    not take over the calling thread's event loop.  When a running asyncio
    loop is found (e.g. inside a Jupyter kernel), browser-initiated viewer
    callbacks are pumped onto that loop so they run on the kernel thread.
    Otherwise they are pumped from a private daemon thread.
    """

    def __init__(self, app, factory,
                 host='localhost', port=9909, ev_quit=None,
                 viewer_class=None):

        self.host = host
        self.port = port
        self.app = app
        self.logger = app.logger
        self.factory = factory
        if ev_quit is None:
            ev_quit = threading.Event()
        self.ev_quit = ev_quit
        self.default_viewer_class = viewer_class

        # callback-pump state
        self._pump_loop = None
        self._pump_handle = None
        self._pump_thread = None
        self._pump_interval = 0.05
        self._pump_stop = threading.Event()

    def start(self, use_thread=True, no_ioloop=True):
        """Start the web application servers.

        Parameters
        ----------
        use_thread : bool
            Unused; retained for backward compatibility.
        no_ioloop : bool
            Retained for backward compatibility.  The servers always run
            in their own background threads now, so this is effectively a
            no-op (kept so existing notebooks calling
            ``server.start(no_ioloop=True)`` keep working).
        """
        # starts the WebSocket server (background thread) and the HTTP
        # file server (background thread)
        self.app.start()

        # keep browser-initiated viewer callbacks flowing
        self._install_pump()

        self.logger.info("Open %s in a browser to view a viewer." %
                         (self.app.get_url(),))

    def _install_pump(self):
        # Find a running event loop to attach the pump to (the Jupyter
        # kernel keeps one running).  If found, callbacks run on that
        # loop's thread; otherwise fall back to a private daemon thread.
        loop = None
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            try:
                loop = asyncio.get_event_loop()
            except RuntimeError:
                loop = None

        self._pump_stop.clear()

        if loop is not None and loop.is_running():
            self._pump_loop = loop
            loop.call_soon_threadsafe(self._pump_once)
            self.logger.debug("pumping viewer events on the running "
                              "asyncio loop")
        else:
            self._pump_loop = None
            self._pump_thread = threading.Thread(target=self._pump_thread_loop,
                                                 daemon=True)
            self._pump_thread.start()
            self.logger.debug("pumping viewer events on a daemon thread")

    def _pump_once(self):
        # runs on the asyncio loop thread; drain pending callbacks without
        # blocking, then reschedule
        if self._pump_stop.is_set():
            return
        try:
            self.app.process_events(0)
        except Exception:
            self.logger.error("error processing viewer events",
                              exc_info=True)
        if self._pump_loop is not None and not self._pump_stop.is_set():
            self._pump_handle = self._pump_loop.call_later(
                self._pump_interval, self._pump_once)

    def _pump_thread_loop(self):
        while not self._pump_stop.is_set():
            try:
                self.app.process_events(self._pump_interval)
            except Exception:
                self.logger.error("error processing viewer events",
                                  exc_info=True)

    def stop(self):
        # stop the callback pump
        self._pump_stop.set()
        if self._pump_loop is not None and self._pump_handle is not None:
            try:
                self._pump_loop.call_soon_threadsafe(self._pump_handle.cancel)
            except Exception:
                pass
        self._pump_loop = None

        # close all sessions and stop the servers
        try:
            self.app.close()
        except Exception:
            self.logger.error("error closing application", exc_info=True)

        self.ev_quit.set()

    def get_viewer(self, v_id, viewer_class=None, width=512, height=512,
                   force_new=False):

        if viewer_class is None:
            viewer_class = self.default_viewer_class

        v_info = self.factory.get_viewer(v_id, viewer_class=viewer_class,
                                         width=width, height=height,
                                         force_new=force_new)
        return v_info.viewer


def make_server(host='localhost', port=9909, ev_quit=None,
                viewer_class=None, logger=None, numthreads=5,
                max_sessions=None, use_opencv=False):
    """Create and return a :class:`WebServer` hosting Ginga web viewers.

    Parameters
    ----------
    host : str
        Interface to bind to (default ``'localhost'``).
    port : int
        HTTP port to serve the viewer pages on (default ``9909``).  The
        WebSocket server uses ``port + 1``.
    ev_quit : threading.Event or None
        Optional shared quit event.
    viewer_class : class or None
        Default viewer class for :meth:`WebServer.get_viewer`
        (defaults to :class:`EnhancedCanvasView`).
    logger : logging-compatible logger or None
        If None, a simple stderr logger is created so the connect URL is
        printed.
    numthreads : int
        Unused; retained for backward compatibility.
    max_sessions : int or None
        Maximum number of concurrent browser sessions.  ``None`` (the
        default) means unlimited, which allows several independent
        viewers.
    use_opencv : bool
        Unused; retained for backward compatibility.

    Returns
    -------
    server : WebServer
        The (not-yet-started) web server.  Call :meth:`WebServer.start`.
    """
    if logger is None:
        logger = log.get_logger("ipg", level=logging.INFO, log_stderr=True)

    # The Ginga pgw Application starts both the WebSocket server and a
    # built-in HTTP server (serving the pgwidgets JS/CSS) -- enable the
    # latter so a browser can load the viewer page.
    app = Widgets.Application(logger=logger, host=host, port=port,
                              http_server=True, max_sessions=max_sessions)

    factory = ViewerFactory(logger, app)

    server = WebServer(app, factory, host=host, port=port,
                       ev_quit=ev_quit, viewer_class=viewer_class)
    return server



#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""Localization support for ginga and applications built on it.

ginga registers its own message domain on import.  A downstream application
adds its own translations by calling :func:`register` once at startup::

    from ginga.locale import localize
    localize.register('myapp', os.path.join(mypkg_dir, 'locale'))

After that, ``_tr()`` (and the caption translation in ``build_info``, plugin
docstring lookups, etc.) consult the application's catalog as well as ginga's,
so the application's own UI strings and help documents are localized through
the same shared widget/plugin machinery.  The active language is global and
applies to every registered catalog.
"""
import gettext
import os

# ginga's own locale directory (kept public for backward compatibility)
localedir = os.path.abspath(os.path.dirname(__file__))

# Registered translation sources, in registration order.  ginga registers
# itself on import; downstream applications call register() to add their own.
# _tr() consults them most-recently-registered first, so an application's
# catalog overrides ginga's for a shared message id (app-catalog-wins).
_registry = []

# The requested language: None/'' honors the environment
# (LANGUAGE/LC_ALL/LC_MESSAGES/LANG); a code overrides it.  Applied to every
# registered catalog.
_lang_pref = None


def _load_catalog(domain, ldir, code):
    languages = None if code in (None, '') else [code]
    # fallback=True: untranslated strings come back unchanged, so English (the
    # source language) always works even with no compiled catalog.
    return gettext.translation(domain, ldir, languages=languages,
                               fallback=True)


def register(domain, localedir):
    """Register a gettext message domain and its locale directory.

    ``localedir`` should contain ``<lang>/LC_MESSAGES/<domain>.mo`` catalogs
    and, optionally, ``<lang>/docs/<category>/*.rst`` help documents.
    Applications built on ginga call this once at startup so their own UI
    strings (including captions passed through ginga's shared widgets) and
    help documents are localized alongside ginga's.  Registering the same
    ``(domain, localedir)`` twice is a no-op.
    """
    for ent in _registry:
        if ent['domain'] == domain and ent['localedir'] == localedir:
            return
    _registry.append(dict(domain=domain, localedir=localedir,
                          catalog=_load_catalog(domain, localedir, _lang_pref)))


# ginga registers itself
register('ginga', localedir)


def _tr(message):
    """Translate `message` into the active language.

    Consults every registered catalog, most-recently-registered first
    (app-catalog-wins); returns the original string if none has a translation.
    """
    for ent in reversed(_registry):
        translated = ent['catalog'].gettext(message)
        if translated != message:
            return translated
    return message


# conventional gettext alias
_ = _tr


def N_(message):
    """Mark `message` for translation without translating it now (gettext_noop).

    Use for strings that must be defined before the UI language is applied
    (e.g. class-attribute lists), so the extractor still records them; then
    translate at display time with ``_tr()``.
    """
    return message


def set_language(code=None):
    """Set the active UI language for all registered catalogs.

    `code` is a locale code such as ``'ja'`` or ``'en'``.  Pass ``None`` or
    ``''`` (the default) to honor the environment variables
    (LANGUAGE/LC_ALL/LC_MESSAGES/LANG).
    """
    global _lang_pref
    _lang_pref = None if code in (None, '') else code
    for ent in _registry:
        ent['catalog'] = _load_catalog(ent['domain'], ent['localedir'],
                                       _lang_pref)


def get_language():
    """Return the code of the active language (e.g. ``'ja'``), or None when
    running in English / fallback.
    """
    if _lang_pref not in (None, ''):
        return _lang_pref
    # env-resolved: read the language from any loaded catalog's header
    for ent in _registry:
        lang = ent['catalog'].info().get('language')
        if lang:
            return lang
    return None


def get_localized_doc(name, category='plugins'):
    """Return a translated long-form help document for the active language,
    or None if there isn't one.

    Searches each registered locale directory for
    ``<localedir>/<lang>/docs/<category>/<name>.rst`` (most-recently-registered
    first).  These are tracked source files (whole-document translations), used
    for text too large for the gettext catalogs (e.g. plugin docstrings shown
    as help).  Returns None (caller uses the English source) when the language
    is English/unset or no file is found.
    """
    lang = get_language()
    if lang is None or lang == 'en':
        return None
    for ent in reversed(_registry):
        path = os.path.join(ent['localedir'], lang, 'docs', category,
                            name + '.rst')
        try:
            with open(path, encoding='utf-8') as f:
                return f.read()
        except OSError:
            continue
    return None


def get_message(key, default='', category='messages'):
    """Return a localized multi-line message block for the active language.

    Like :func:`get_localized_doc`, but for paragraph-sized help text that
    lives inline in the source: `default` (the English text) is returned when
    there is no translation file
    (``<localedir>/<lang>/docs/<category>/<key>.rst``).
    """
    text = get_localized_doc(key, category=category)
    return default if text is None else text


# All widget-type tokens understood by build_info() captions.  Used by the
# custom Babel extractor to recognize a caption row.
KNOWN_WTYPES = frozenset([
    'label', 'llabel', 'textentry', 'entry', 'textentryset', 'entryset',
    'combobox', 'comboboxedit', 'spinbox', 'spinbutton', 'spinfloat',
    'vbox', 'hbox', 'hslider', 'hscale', 'vslider', 'vscale',
    'checkbox', 'checkbutton', 'radiobutton', 'togglebutton', 'button',
    'spacer', 'textarea', 'toolbar', 'progress', 'menubar', 'dial',
])

# The subset of caption widget types whose title is shown to the user and so
# should be translated.  (llabel is excluded: its text is a dynamic value,
# set at runtime, not a fixed label.)
TRANSLATABLE_WTYPES = frozenset([
    'label', 'button', 'checkbox', 'checkbutton', 'radiobutton',
    'togglebutton',
])

# Prefix that opts a caption title out of translation (and is stripped).
NO_TRANSLATE_PREFIX = '^'


def translate_caption(title, wtype):
    """Process one build_info() caption ``(title, wtype)`` pair.

    Returns ``(key_title, disp_title)``:

    - ``key_title`` is the English string from which the Bunch key is derived
      (with any leading ``^`` opt-out marker stripped).  Deriving the key from
      English keeps widget keys stable across languages so existing code that
      looks up ``b.<name>`` keeps working.
    - ``disp_title`` is the text actually shown: translated when the widget
      type displays its title and the title is not opted out with a leading
      ``^``; otherwise the (marker-stripped) English text.
    """
    if title.startswith(NO_TRANSLATE_PREFIX):
        title = title[len(NO_TRANSLATE_PREFIX):]
        return title, title
    if wtype in TRANSLATABLE_WTYPES:
        return title, _tr(title)
    return title, title


# Display names for the shipped languages, each written in its own script.
# Used for the Language menu; these are shown as-is regardless of the active
# UI language (a language is self-identifying in its own script).
LANGUAGE_NAMES = {
    'en': 'English',
    'de': 'Deutsch',
    'es': 'Español',
    'fr': 'Français',
    'it': 'Italiano',
    'ja': '日本語',
    'ko': '한국어',
    'zh': '中文',
}


def get_language_name(code):
    """Return the native-script display name for a language code, or the code
    itself if unknown.
    """
    return LANGUAGE_NAMES.get(code, code)


def get_available_languages():
    """Return the sorted list of language codes with a compiled catalog
    (``<code>/LC_MESSAGES/<domain>.mo``) installed in any registered locale
    directory.  English (the source language) is always included, since it
    works via fallback.
    """
    langs = set(['en'])
    for ent in _registry:
        try:
            for name in os.listdir(ent['localedir']):
                mo = os.path.join(ent['localedir'], name, 'LC_MESSAGES',
                                  ent['domain'] + '.mo')
                if os.path.isfile(mo):
                    langs.add(name)
        except OSError:
            pass
    return sorted(langs)
